# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import types
from collections.abc import Callable, Iterator
from typing import (
    Any,
    Generic,
    Literal,
    Protocol,
    Sequence,
    Type,
    TypeGuard,
    TypeVar,
    cast,
    overload,
)

import numpy as np

from gt4py.eve import extended_typing as xtyping, utils
from gt4py.next import common
from gt4py.next.iterator.type_system import type_specifications as it_ts
from gt4py.next.type_system import type_specifications as ts


def _number_to_ordinal_number(number: int) -> str:
    """
    Convert number into ordinal number.

    >>> for i in range(0, 5):
    ...     print(_number_to_ordinal_number(i))
    0th
    1st
    2nd
    3rd
    4th
    """
    number_as_string = str(number)
    if len(number_as_string) > 1 and number_as_string[-2] == "1":
        return number_as_string + "th"
    last_digit = number_as_string[-1]
    if last_digit == "1":
        return number_as_string + "st"
    if last_digit == "2":
        return number_as_string + "nd"
    if last_digit == "3":
        return number_as_string + "rd"
    return number_as_string + "th"


def is_concrete(symbol_type: ts.TypeSpec) -> TypeGuard[ts.TypeSpec]:
    """Figure out if the foast type is completely deduced."""
    if isinstance(symbol_type, ts.DeferredType):
        return False
    elif isinstance(symbol_type, ts.TypeSpec):
        return True
    return False


def type_class(symbol_type: ts.TypeSpec) -> Type[ts.TypeSpec]:
    """
    Determine which class should be used to create a compatible concrete type.

    Examples:
    ---------
    >>> type_class(ts.DeferredType(constraint=ts.ScalarType)).__name__
    'ScalarType'

    >>> type_class(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.BOOL))).__name__
    'FieldType'

    >>> type_class(ts.TupleType(types=[])).__name__
    'TupleType'
    """
    if isinstance(symbol_type, ts.DeferredType):
        constraint = symbol_type.constraint
        if constraint is None:
            raise ValueError(f"No type information available for '{symbol_type}'.")
        elif isinstance(constraint, tuple):
            raise ValueError(f"Not sufficient type information available for '{symbol_type}'.")
        return constraint
    if isinstance(symbol_type, ts.TypeSpec):
        return symbol_type.__class__
    raise ValueError(
        f"Invalid type for TypeInfo: requires '{ts.TypeSpec}', got '{type(symbol_type)}'."
    )


@overload
def primitive_constituents(
    symbol_type: ts.TypeSpec, with_path_arg: Literal[False] = False
) -> utils.XIterable[ts.TypeSpec]: ...


@overload
def primitive_constituents(
    symbol_type: ts.TypeSpec, with_path_arg: Literal[True]
) -> utils.XIterable[tuple[ts.TypeSpec, tuple[int, ...]]]: ...


def primitive_constituents(
    symbol_type: ts.TypeSpec, with_path_arg: bool = False
) -> utils.XIterable[ts.TypeSpec] | utils.XIterable[tuple[ts.TypeSpec, tuple[int, ...]]]:
    """
    Return the primitive types contained in a composite type.

    >>> from gt4py.next import common
    >>> I = common.Dimension(value="I")
    >>> int_type = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> field_type = ts.FieldType(dims=[I], dtype=int_type)

    >>> tuple_type = ts.TupleType(types=[int_type, field_type])
    >>> primitive_constituents(tuple_type).to_list()  # doctest: +ELLIPSIS
    [ScalarType(...), FieldType(...)]

    >>> nested_tuple = ts.TupleType(types=[field_type, tuple_type])
    >>> primitive_constituents(nested_tuple).to_list()  # doctest: +ELLIPSIS
    [FieldType(...), ScalarType(...), FieldType(...)]
    """

    def constituents_yielder(
        symbol_type: ts.TypeSpec, path: tuple[int, ...]
    ) -> Iterator[ts.TypeSpec] | Iterator[tuple[ts.TypeSpec, tuple[int, ...]]]:
        if isinstance(symbol_type, ts.COLLECTION_TYPE_SPECS):
            symbol_type = cast(
                ts.CollectionTypeSpec, symbol_type
            )  # This shouldn't be needed after the previous isinstance() check
            for i, el_type in enumerate(symbol_type.types):
                yield from constituents_yielder(el_type, (*path, i))
        else:
            if with_path_arg:
                yield (symbol_type, path)
            else:
                yield symbol_type

    return utils.xiter(constituents_yielder(symbol_type, ()))  # type: ignore[return-value] # why resolved to XIterable[object]?


_R = TypeVar("_R", covariant=True)
_T = TypeVar("_T")


class TupleConstructorType(Protocol, Generic[_R]):
    def __call__(self, *args: Any) -> _R: ...


def apply_to_primitive_constituents(
    fun: Callable[..., _T],
    *symbol_types: ts.TypeSpec,
    with_path_arg: bool = False,
    tuple_constructor: TupleConstructorType[_R] = lambda *elements: ts.TupleType(types=[*elements]),  # type: ignore[assignment] # probably related to https://github.com/python/mypy/issues/10854
    _path: tuple[int, ...] = (),
) -> _T | _R:
    """
    Apply function to all primitive constituents of a type.

    >>> int_type = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> tuple_type = ts.TupleType(types=[int_type, int_type])
    >>> print(
    ...     apply_to_primitive_constituents(
    ...         lambda primitive_type: ts.FieldType(dims=[], dtype=primitive_type),
    ...         tuple_type,
    ...     )
    ... )
    tuple[Field[[], int64], Field[[], int64]]

    >>> apply_to_primitive_constituents(
    ...     lambda primitive_type, path: (path, primitive_type),
    ...     tuple_type,
    ...     with_path_arg=True,
    ...     tuple_constructor=lambda *elements: dict(elements),
    ... )
    {(0,): ScalarType(kind=<ScalarKind.INT64: 8>, shape=None), (1,): ScalarType(kind=<ScalarKind.INT64: 8>, shape=None)}
    """
    if isinstance(symbol_types[0], ts.TupleType):
        assert all(isinstance(symbol_type, ts.TupleType) for symbol_type in symbol_types)

        return tuple_constructor(
            *[
                apply_to_primitive_constituents(
                    fun,
                    *el_types,
                    _path=(*_path, i),
                    with_path_arg=with_path_arg,
                    tuple_constructor=tuple_constructor,
                )
                for i, el_types in enumerate(
                    zip(*(symbol_type.types for symbol_type in symbol_types))  # type: ignore[attr-defined]  # ensured by assert above
                )
            ]
        )
    if with_path_arg:
        return fun(*symbol_types, path=_path)
    else:
        return fun(*symbol_types)


def extract_dtype(symbol_type: ts.TypeSpec) -> ts.ScalarType | ts.ListType:
    """
    Extract the data type from ``symbol_type`` if it is either `FieldType` or `ScalarType`.

    Raise an error if no dtype can be found or the result would be ambiguous.

    Examples:
    ---------
    >>> print(extract_dtype(ts.ScalarType(kind=ts.ScalarKind.FLOAT64)))
    float64

    >>> print(extract_dtype(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.BOOL))))
    bool
    """
    match symbol_type:
        case ts.FieldType(dtype=dtype):
            return dtype
        case ts.ScalarType() as dtype:
            return dtype
    raise ValueError(f"Can not unambiguosly extract data type from '{symbol_type}'.")


def is_floating_point(symbol_type: ts.TypeSpec) -> bool:
    """
    Check if the dtype of ``symbol_type`` is a floating point type.

    Examples:
    ---------
    >>> is_floating_point(ts.ScalarType(kind=ts.ScalarKind.FLOAT64))
    True
    >>> is_floating_point(ts.ScalarType(kind=ts.ScalarKind.FLOAT32))
    True
    >>> is_floating_point(ts.ScalarType(kind=ts.ScalarKind.INT32))
    False
    >>> is_floating_point(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.FLOAT32)))
    True
    """
    if not isinstance(symbol_type, (ts.ScalarType, ts.FieldType)):
        return False
    return isinstance(dtype := extract_dtype(symbol_type), ts.ScalarType) and dtype.kind in [
        ts.ScalarKind.FLOAT32,
        ts.ScalarKind.FLOAT64,
    ]


def is_integer(symbol_type: ts.TypeSpec) -> bool:
    """
    Check if ``symbol_type`` is an integral type.

    Examples:
    ---------
    >>> is_integer(ts.ScalarType(kind=ts.ScalarKind.INT32))
    True
    >>> is_integer(ts.ScalarType(kind=ts.ScalarKind.FLOAT32))
    False
    >>> is_integer(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.INT32)))
    False
    """
    return isinstance(symbol_type, ts.ScalarType) and symbol_type.kind in {
        ts.ScalarKind.INT8,
        ts.ScalarKind.UINT8,
        ts.ScalarKind.INT16,
        ts.ScalarKind.UINT16,
        ts.ScalarKind.INT32,
        ts.ScalarKind.UINT32,
        ts.ScalarKind.INT64,
    }


def is_integral(symbol_type: ts.TypeSpec) -> bool:
    """
    Check if the dtype of ``symbol_type`` is an integral type.

    Examples:
    ---------
    >>> is_integral(ts.ScalarType(kind=ts.ScalarKind.INT32))
    True
    >>> is_integral(ts.ScalarType(kind=ts.ScalarKind.FLOAT32))
    False
    >>> is_integral(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.INT32)))
    True
    """
    if not isinstance(symbol_type, (ts.ScalarType, ts.FieldType)):
        return False
    return is_integer(extract_dtype(symbol_type))


def is_number(symbol_type: ts.TypeSpec) -> bool:
    """
    Check if ``symbol_type`` is either intergral or float.

    Examples:
    ---------
    >>> is_number(ts.ScalarType(kind=ts.ScalarKind.FLOAT64))
    True
    >>> is_number(ts.ScalarType(kind=ts.ScalarKind.INT32))
    True
    >>> is_number(ts.ScalarType(kind=ts.ScalarKind.BOOL))
    False
    >>> is_number(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.INT64)))
    False
    """
    if not isinstance(symbol_type, ts.ScalarType):
        return False
    # TODO(nfarabullini): re-factor is_arithmetic such that it only checks for scalars
    #  and the emtpy field pass in an another function
    return is_arithmetic(symbol_type)


def is_logical(symbol_type: ts.TypeSpec) -> bool:
    return (
        isinstance(symbol_type, (ts.FieldType, ts.ScalarType))
        and isinstance(dtype := extract_dtype(symbol_type), ts.ScalarType)
        and dtype.kind is ts.ScalarKind.BOOL
    )


def is_arithmetic(symbol_type: ts.TypeSpec) -> bool:
    """
    Check if ``symbol_type`` is compatible with arithmetic operations.

    Examples:
    ---------
    >>> is_arithmetic(ts.ScalarType(kind=ts.ScalarKind.FLOAT64))
    True
    >>> is_arithmetic(ts.ScalarType(kind=ts.ScalarKind.BOOL))
    False
    >>> is_arithmetic(ts.ScalarType(kind=ts.ScalarKind.STRING))
    False
    >>> is_arithmetic(ts.FieldType(dims=[], dtype=ts.ScalarType(kind=ts.ScalarKind.INT32)))
    True
    """
    return is_floating_point(symbol_type) or is_integral(symbol_type)


def arithmetic_bounds(arithmetic_type: ts.ScalarType) -> tuple[np.number, np.number]:
    assert is_arithmetic(arithmetic_type)
    return {  # type: ignore[return-value] # why resolved to `tuple[object, object]`?
        ts.ScalarKind.FLOAT32: (np.finfo(np.float32).min, np.finfo(np.float32).max),
        ts.ScalarKind.FLOAT64: (np.finfo(np.float64).min, np.finfo(np.float64).max),
        ts.ScalarKind.INT8: (np.iinfo(np.int8).min, np.iinfo(np.int8).max),
        ts.ScalarKind.UINT8: (np.iinfo(np.uint8).min, np.iinfo(np.uint8).max),
        ts.ScalarKind.INT16: (np.iinfo(np.int16).min, np.iinfo(np.int16).max),
        ts.ScalarKind.UINT16: (np.iinfo(np.uint16).min, np.iinfo(np.uint16).max),
        ts.ScalarKind.INT32: (np.iinfo(np.int32).min, np.iinfo(np.int32).max),
        ts.ScalarKind.UINT32: (np.iinfo(np.uint32).min, np.iinfo(np.uint32).max),
        ts.ScalarKind.INT64: (np.iinfo(np.int64).min, np.iinfo(np.int64).max),
        ts.ScalarKind.UINT64: (np.iinfo(np.uint64).min, np.iinfo(np.uint64).max),
    }[arithmetic_type.kind]


def is_type_or_tuple_of_type(type_: ts.TypeSpec, expected_type: type | tuple) -> bool:
    """
    Return True if ``type_`` matches any of the expected or is a tuple of them.

    Examples:
    ---------
    >>> scalar_type = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> field_type = ts.FieldType(dims=[], dtype=scalar_type)
    >>> is_type_or_tuple_of_type(field_type, ts.FieldType)
    True
    >>> is_type_or_tuple_of_type(
    ...     ts.TupleType(types=[scalar_type, field_type]), (ts.ScalarType, ts.FieldType)
    ... )
    True
    >>> is_type_or_tuple_of_type(scalar_type, ts.FieldType)
    False
    >>> is_type_or_tuple_of_type(ts.TupleType(types=[scalar_type, field_type]), ts.FieldType)
    False
    """
    return all(isinstance(t, expected_type) for t in primitive_constituents(type_))


def is_tuple_of_type(type_: ts.TypeSpec, expected_type: type | tuple) -> TypeGuard[ts.TupleType]:
    """
    Return True if ``type_`` matches (nested) tuple of ``expected_type``.

    Examples:
    ---------
    >>> scalar_type = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> field_type = ts.FieldType(dims=[], dtype=scalar_type)
    >>> is_tuple_of_type(field_type, ts.FieldType)
    False
    >>> is_tuple_of_type(
    ...     ts.TupleType(types=[scalar_type, field_type]), (ts.ScalarType, ts.FieldType)
    ... )
    True
    >>> is_tuple_of_type(ts.TupleType(types=[scalar_type]), ts.FieldType)
    False
    >>> is_tuple_of_type(ts.TupleType(types=[scalar_type, field_type]), ts.FieldType)
    False
    """
    return isinstance(type_, ts.TupleType) and is_type_or_tuple_of_type(type_, expected_type)


def extract_dims(symbol_type: ts.TypeSpec) -> list[common.Dimension]:
    """
    Try to extract field dimensions if possible.

    Scalars are treated as zero-dimensional

    Examples:
    ---------
    >>> extract_dims(ts.ScalarType(kind=ts.ScalarKind.INT64, shape=[3, 4]))
    []
    >>> I = common.Dimension(value="I")
    >>> J = common.Dimension(value="J")
    >>> extract_dims(ts.FieldType(dims=[I, J], dtype=ts.ScalarType(kind=ts.ScalarKind.INT64)))
    [Dimension(value='I', kind=<DimensionKind.HORIZONTAL: 'horizontal'>), Dimension(value='J', kind=<DimensionKind.HORIZONTAL: 'horizontal'>)]
    """
    if isinstance(symbol_type, ts.ScalarType):
        return []
    if isinstance(symbol_type, ts.FieldType):
        return symbol_type.dims
    raise ValueError(f"Can not extract dimensions from '{symbol_type}'.")


def is_local_field(type_: ts.FieldType) -> bool:
    """
    Return if `type_` is a field defined on a local dimension.

    Examples:
    ---------
    >>> V = common.Dimension(value="V")
    >>> V2E = common.Dimension(value="V2E", kind=common.DimensionKind.LOCAL)
    >>> is_local_field(ts.FieldType(dims=[V, V2E], dtype=ts.ScalarType(kind=ts.ScalarKind.INT64)))
    True
    >>> is_local_field(ts.FieldType(dims=[V], dtype=ts.ScalarType(kind=ts.ScalarKind.INT64)))
    False
    """
    return any(dim.kind == common.DimensionKind.LOCAL for dim in type_.dims)


def contains_local_field(type_: ts.TypeSpec) -> bool:
    """Return if primitive constitutens of `type_` contains a field defined on a local dimension."""
    return any(
        isinstance(t, ts.FieldType) and is_local_field(t) for t in primitive_constituents(type_)
    )


# TODO(tehrengruber): This function has specializations on Iterator types, which are not part of
#  the general / shared type system. This functionality should be moved to the iterator-only
#  type system, but we need some sort of multiple dispatch for that.
# TODO(tehrengruber): Should this have a direction like is_concretizable?
def is_compatible_type(type_a: ts.TypeSpec, type_b: ts.TypeSpec) -> bool:
    """
    Predicate to determine if two types are compatible.

    This function gracefully handles:
     - iterators with unknown positions which are considered compatible to any other positions
       of another iterator.
     - iterators which are defined everywhere, i.e. empty defined dimensions
    Beside that this function simply checks for equality of types.

    >>> bool_type = ts.ScalarType(kind=ts.ScalarKind.BOOL)
    >>> IDim = common.Dimension(value="IDim")
    >>> type_on_i_of_i_it = it_ts.IteratorType(
    ...     position_dims=[IDim], defined_dims=[IDim], element_type=bool_type
    ... )
    >>> type_on_undefined_of_i_it = it_ts.IteratorType(
    ...     position_dims="unknown", defined_dims=[IDim], element_type=bool_type
    ... )
    >>> is_compatible_type(type_on_i_of_i_it, type_on_undefined_of_i_it)
    True

    >>> JDim = common.Dimension(value="JDim")
    >>> type_on_j_of_j_it = it_ts.IteratorType(
    ...     position_dims=[JDim], defined_dims=[JDim], element_type=bool_type
    ... )
    >>> is_compatible_type(type_on_i_of_i_it, type_on_j_of_j_it)
    False
    """
    is_compatible = True

    if isinstance(type_a, it_ts.IteratorType) and isinstance(type_b, it_ts.IteratorType):
        if not any(el_type.position_dims == "unknown" for el_type in [type_a, type_b]):
            is_compatible &= type_a.position_dims == type_b.position_dims
        if type_a.defined_dims and type_b.defined_dims:
            is_compatible &= type_a.defined_dims == type_b.defined_dims
        is_compatible &= type_a.element_type == type_b.element_type
    elif isinstance(type_a, ts.TupleType) and isinstance(type_b, ts.TupleType):
        if len(type_a.types) != len(type_b.types):
            return False
        for el_type_a, el_type_b in zip(type_a.types, type_b.types, strict=True):
            is_compatible &= is_compatible_type(el_type_a, el_type_b)
    elif isinstance(type_a, ts.NamedCollectionType) and isinstance(type_b, ts.NamedCollectionType):
        if type_a.keys != type_b.keys:
            return False
        if (
            not any(
                python_type is ts.ANY_PYTHON_TYPE_NAME
                for python_type in [type_a.original_python_type, type_b.original_python_type]
            )
            and type_a.original_python_type != type_b.original_python_type
        ):
            return False
        for el_type_a, el_type_b in zip(type_a.types, type_b.types, strict=True):
            is_compatible &= is_compatible_type(el_type_a, el_type_b)
    elif isinstance(type_a, ts.FunctionType) and isinstance(type_b, ts.FunctionType):
        for arg_a, arg_b in zip(type_a.pos_only_args, type_b.pos_only_args, strict=True):
            is_compatible &= is_compatible_type(arg_a, arg_b)
        for arg_a, arg_b in zip(
            type_a.pos_or_kw_args.values(), type_b.pos_or_kw_args.values(), strict=True
        ):
            is_compatible &= is_compatible_type(arg_a, arg_b)
        for arg_a, arg_b in zip(
            type_a.kw_only_args.values(), type_b.kw_only_args.values(), strict=True
        ):
            is_compatible &= is_compatible_type(arg_a, arg_b)
        is_compatible &= is_compatible_type(type_a.returns, type_b.returns)
    else:
        is_compatible &= is_concretizable(type_a, type_b)

    return is_compatible


def is_concretizable(symbol_type: ts.TypeSpec, to_type: ts.TypeSpec) -> bool:
    """
    Check if ``symbol_type`` can be concretized to ``to_type``.

    Examples:
    ---------
    >>> is_concretizable(
    ...     ts.ScalarType(kind=ts.ScalarKind.INT64), to_type=ts.ScalarType(kind=ts.ScalarKind.INT64)
    ... )
    True

    >>> is_concretizable(
    ...     ts.ScalarType(kind=ts.ScalarKind.INT64),
    ...     to_type=ts.ScalarType(kind=ts.ScalarKind.FLOAT64),
    ... )
    False

    >>> is_concretizable(
    ...     ts.DeferredType(constraint=None),
    ...     to_type=ts.FieldType(dtype=ts.ScalarType(kind=ts.ScalarKind.BOOL), dims=[]),
    ... )
    True

    >>> is_concretizable(
    ...     ts.DeferredType(constraint=ts.DataType),
    ...     to_type=ts.FieldType(dtype=ts.ScalarType(kind=ts.ScalarKind.BOOL), dims=[]),
    ... )
    True

    >>> is_concretizable(
    ...     ts.DeferredType(constraint=ts.OffsetType),
    ...     to_type=ts.FieldType(dtype=ts.ScalarType(kind=ts.ScalarKind.BOOL), dims=[]),
    ... )
    False

    >>> is_concretizable(
    ...     ts.DeferredType(constraint=ts.TypeSpec),
    ...     to_type=ts.DeferredType(constraint=ts.ScalarType),
    ... )
    True

    """
    if isinstance(symbol_type, ts.DeferredType) and (
        symbol_type.constraint is None
        or (isinstance(to_type, ts.DeferredType) and to_type.constraint is None)
        or issubclass(type_class(to_type), symbol_type.constraint)
    ):
        return True
    elif is_concrete(symbol_type):
        return symbol_type == to_type
    return False


def promote(
    *types: ts.FieldType | ts.ScalarType, always_field: bool = False
) -> ts.FieldType | ts.ScalarType:
    """
    Promote a set of field or scalar types to a common type.

    The resulting type is defined on all dimensions of the arguments, respecting
    the individual order of the dimensions of each argument (see
    :func:`common.promote_dims` for more details).

    >>> dtype = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> I, J, K = (common.Dimension(value=dim) for dim in ["I", "J", "K"])
    >>> promoted: ts.FieldType = promote(
    ...     ts.FieldType(dims=[I, J], dtype=dtype), ts.FieldType(dims=[I, J, K], dtype=dtype), dtype
    ... )
    >>> promoted.dims == [I, J, K] and promoted.dtype == dtype
    True

    >>> promoted: ts.FieldType = promote(
    ...     ts.FieldType(dims=[I, J], dtype=dtype), ts.FieldType(dims=[K], dtype=dtype)
    ... )
    >>> promoted.dims == [I, J, K] and promoted.dtype == dtype
    True
    """
    if not always_field and all(isinstance(type_, ts.ScalarType) for type_ in types):
        if not all(type_ == types[0] for type_ in types):
            raise ValueError("Could not promote scalars of different dtype (not implemented).")
        if not all(type_.shape is None for type_ in types):  # type: ignore[union-attr]
            raise NotImplementedError("Shape promotion not implemented.")
        return types[0]
    elif all(isinstance(type_, (ts.ScalarType, ts.FieldType)) for type_ in types):
        dims = common.promote_dims(*(extract_dims(type_) for type_ in types))
        extracted_dtypes = [extract_dtype(type_) for type_ in types]
        assert all(isinstance(dtype, ts.ScalarType) for dtype in extracted_dtypes)
        dtype = cast(ts.ScalarType, promote(*extracted_dtypes))  # type: ignore[arg-type] # checked is `ScalarType`

        return ts.FieldType(dims=dims, dtype=dtype)
    raise TypeError("Expected a 'FieldType' or 'ScalarType'.")


@functools.singledispatch
def return_type(
    callable_type: ts.CallableType,
    *,
    with_args: Sequence[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.TypeSpec:
    raise NotImplementedError(
        f"Return type deduction of type '{type(callable_type).__name__}' not implemented."
    )


@return_type.register
def return_type_func(
    func_type: ts.FunctionType,
    *,
    with_args: Sequence[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.TypeSpec:
    return func_type.returns


@return_type.register
def return_type_field(
    field_type: ts.FieldType,
    *,
    with_args: Sequence[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.FieldType:
    try:
        accepts_args(
            field_type,
            with_args=with_args,
            with_kwargs=with_kwargs,
            raise_exception=True,
        )
    except ValueError as ex:
        raise ValueError("Could not deduce return type of invalid remap operation.") from ex

    if not isinstance(with_args[0], ts.OffsetType):
        raise ValueError(f"First argument must be of type '{ts.OffsetType}', got '{with_args[0]}'.")

    source_dim = with_args[0].source
    target_dims = with_args[0].target
    new_dims = []
    # TODO: This code does not handle ellipses for dimensions. Fix it.
    assert field_type.dims is not ...
    for d in field_type.dims:
        if d != source_dim:
            new_dims.append(d)
        else:
            new_dims.extend(target_dims)
    return ts.FieldType(dims=new_dims, dtype=field_type.dtype)


@return_type.register
def return_type_constructor(
    constructor_type: ts.ConstructorType,
    *,
    with_args: Sequence[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
) -> ts.TypeSpec:
    return constructor_type.definition.returns


UNDEFINED_ARG = types.new_class("UNDEFINED_ARG")


# TODO(egparedes): replace by `inspect.Signature.bind()`
@functools.singledispatch
def canonicalize_arguments(
    func_type: ts.CallableType,
    args: Sequence,
    kwargs: dict,
    *,
    ignore_errors: bool = False,
) -> tuple[tuple, dict]:
    raise NotImplementedError(f"Not implemented for type '{type(func_type).__name__}'.")


@canonicalize_arguments.register
def canonicalize_function_arguments(
    func_type: ts.FunctionType,
    args: Sequence,
    kwargs: dict,
    *,
    ignore_errors: bool = False,
) -> tuple[tuple, dict]:
    num_pos_only_args = len(func_type.pos_only_args)
    num_pos_or_kw_args = len(func_type.pos_or_kw_args)
    canonical_args = [*args] + (
        [UNDEFINED_ARG] * max(num_pos_only_args + num_pos_or_kw_args - len(args), 0)
    )
    remaining_kwargs = {**kwargs}

    pos_or_kw_args_names = [*func_type.pos_or_kw_args]
    for name in kwargs:
        if name in func_type.pos_or_kw_args:
            args_idx = num_pos_only_args + pos_or_kw_args_names.index(name)
            if canonical_args[args_idx] is UNDEFINED_ARG:
                canonical_args[args_idx] = remaining_kwargs.pop(name)
            elif not ignore_errors:
                raise ValueError(
                    f"Error canonicalizing function arguments. Got multiple values for argument '{name}'."
                )

    missing_kw_args = func_type.kw_only_args.keys() - remaining_kwargs.keys()
    invalid_kw_args = remaining_kwargs.keys() - func_type.kw_only_args.keys()

    if not ignore_errors:
        if missing_kw_args:
            raise ValueError(f"Missing required keyword arguments: {[*missing_kw_args]}.")
        if invalid_kw_args:
            raise ValueError(f"Invalid keyword arguments: {[*invalid_kw_args]}.")

    # Sort remaining keyword arguments in the signature ordering,
    # keeping the invalid ones at the end if 'ignore_errors' is True.
    canonical_kwargs = {
        k: remaining_kwargs[k] for k in func_type.kw_only_args if k in remaining_kwargs
    } | {k: remaining_kwargs[k] for k in invalid_kw_args}

    return tuple(canonical_args), canonical_kwargs


@canonicalize_arguments.register(ts.ConstructorType)
def canonicalize_constructor_arguments(
    constructor_type: ts.ConstructorType,
    args: tuple | list,
    kwargs: dict,
    *,
    ignore_errors: bool = False,
) -> tuple[tuple, dict]:
    return canonicalize_arguments(
        constructor_type.definition,
        args,
        kwargs,
        ignore_errors=ignore_errors,
    )


def structural_function_signature_incompatibilities(
    func_type: ts.FunctionType, args: Sequence, kwargs: dict[str, Any]
) -> Iterator[str]:
    """
    Return structural incompatibilities for a call to ``func_type`` with given arguments.

    This function requires `args` and `kwargs` to be canonicalized using
    `canonicalize_function_arguments`.

    Contrary to `function_signature_incompatibilities` the arguments don't need to be types, but
    can be anything. This function merely checks whether their structure matches (e.g. for every
    parameter in the signature there is a corresponding argument value).
    """
    kwargs = {**kwargs}

    # check positional arguments
    for name in [*kwargs]:
        if name in func_type.pos_or_kw_args:
            args_idx = len(func_type.pos_only_args) + list(func_type.pos_or_kw_args.keys()).index(
                name
            )
            if args_idx < len(args):
                # remove the argument here such that later errors stay comprehensible
                kwargs.pop(name)
                yield f"Got multiple values for argument '{name}'."

    num_pos_params = len(func_type.pos_only_args) + len(func_type.pos_or_kw_args)
    num_pos_args = len(args) - args.count(UNDEFINED_ARG)
    if num_pos_params != num_pos_args:
        if len(kwargs) > 0:
            kwargs_msg = f"positional argument{'s' if num_pos_params != 1 else ''} (and {len(kwargs)} keyword-only argument{'s' if len(kwargs) != 1 else ''}) "
        else:
            kwargs_msg = ""
        yield f"Function takes {num_pos_params} positional argument{'s' if num_pos_params != 1 else ''}, but {num_pos_args} {kwargs_msg}were given."

    missing_positional_args = [
        f"'{arg_type}'"
        for arg, arg_type in zip(
            args[len(func_type.pos_only_args) : num_pos_params],
            func_type.pos_or_kw_args.keys(),
        )
        if arg is UNDEFINED_ARG
    ]

    if missing_positional_args:
        yield f"Missing {len(missing_positional_args)} required positional argument{'s' if len(missing_positional_args) != 1 else ''}: {', '.join(missing_positional_args)}"

    # check for missing or extra keyword arguments
    kw_a_m_b = func_type.kw_only_args.keys() - kwargs.keys()
    if len(kw_a_m_b) > 0:
        yield f"Missing required keyword argument{'s' if len(kw_a_m_b) != 1 else ''} '{', '.join(kw_a_m_b)}'."
    kw_b_m_a = kwargs.keys() - func_type.kw_only_args.keys()
    if len(kw_b_m_a) > 0:
        yield f"Got unexpected keyword argument{'s' if len(kw_b_m_a) != 1 else ''} '{', '.join(kw_b_m_a)}'."


@functools.singledispatch
def function_signature_incompatibilities(
    func_type: ts.CallableType, args: Sequence[ts.TypeSpec], kwargs: dict[str, ts.TypeSpec]
) -> Iterator[str]:
    """
    Return incompatibilities for a call to ``func_type`` with given arguments.

    Note that all types must be concrete/complete.
    """
    raise NotImplementedError(f"Not implemented for type '{type(func_type).__name__}'.")


@function_signature_incompatibilities.register
def function_signature_incompatibilities_func(
    func_type: ts.FunctionType,
    args: Sequence[ts.TypeSpec],
    kwargs: dict[str, ts.TypeSpec],
    *,
    skip_canonicalization: bool = False,
    skip_structural_checks: bool = False,
) -> Iterator[str]:
    if not skip_canonicalization:
        args, kwargs = canonicalize_arguments(func_type, args, kwargs, ignore_errors=True)

    # check for structural errors (e.g. wrong number of arguments)
    if not skip_structural_checks:
        error_list = list(structural_function_signature_incompatibilities(func_type, args, kwargs))
        if len(error_list) > 0:
            yield from error_list
            return

    num_pos_only_args = len(func_type.pos_only_args)
    num_pos_or_kw_args = len(func_type.pos_or_kw_args)
    num_pos_params = num_pos_only_args + num_pos_or_kw_args
    assert len(args) >= num_pos_params
    for i, (a_arg, b_arg) in enumerate(
        zip(list(func_type.pos_only_args) + list(func_type.pos_or_kw_args.values()), args)
    ):
        if b_arg is not UNDEFINED_ARG and a_arg != b_arg and not is_compatible_type(a_arg, b_arg):
            if i < num_pos_only_args:
                arg_repr = f"{_number_to_ordinal_number(i + 1)} argument"
            else:
                arg_repr = (
                    f"argument '{list(func_type.pos_or_kw_args.keys())[i - num_pos_only_args]}'"
                )
            yield f"Expected {arg_repr} to be of type '{a_arg}', got '{b_arg}'."

    for kwarg in func_type.kw_only_args.keys() & kwargs.keys():
        if (a_kwarg := func_type.kw_only_args[kwarg]) != (
            b_kwarg := kwargs[kwarg]
        ) and not is_compatible_type(a_kwarg, b_kwarg):
            yield f"Expected keyword argument '{kwarg}' to be of type '{func_type.kw_only_args[kwarg]}', got '{kwargs[kwarg]}'."


@function_signature_incompatibilities.register
def function_signature_incompatibilities_field(
    field_type: ts.FieldType,
    args: list[ts.TypeSpec],
    kwargs: dict[str, ts.TypeSpec],
) -> Iterator[str]:
    if len(args) < 1:
        yield f"Function takes at least 1 argument, but {len(args)} were given."
        return
    for arg in args:
        if not isinstance(arg, ts.OffsetType):
            yield f"Expected arguments to be of type '{ts.OffsetType}', got '{arg}'."
            return
        if len(args) > 1 and len(arg.target) > 1:
            yield f"Function takes only 1 argument in unstructured case, but {len(args)} were given."
            return

    if kwargs:
        yield f"Got unexpected keyword argument(s) '{', '.join(kwargs.keys())}'."
        return

    source_dim = args[0].source  # type: ignore[attr-defined] # ensured by loop above
    target_dims = args[0].target  # type: ignore[attr-defined] # ensured by loop above
    assert field_type.dims is not ...
    if field_type.dims and source_dim not in field_type.dims:
        yield (
            f"Incompatible offset can not shift field defined on "
            f"{', '.join([dim.value for dim in field_type.dims])} from "
            f"{source_dim.value} to target dim(s): "
            f"{', '.join([dim.value for dim in target_dims])}"
        )


@function_signature_incompatibilities.register
def function_signature_incompatibilities_constructor(
    constructor_type: ts.ConstructorType,
    args: Sequence[ts.TypeSpec],
    kwargs: dict[str, ts.TypeSpec],
) -> Iterator[str]:
    yield from function_signature_incompatibilities_func(constructor_type.definition, args, kwargs)


# TODO(havogt): Consider inlining the usage of this function in the call sites
# to get rid of the `raise_exception` case and because the error message here is possibly too specific.
def accepts_args(
    callable_type: ts.CallableType,
    *,
    with_args: Sequence[ts.TypeSpec],
    with_kwargs: dict[str, ts.TypeSpec],
    raise_exception: bool = False,
) -> bool:
    """
    Check if a function can be called for given arguments.

    If ``raise_exception`` is given a :class:`ValueError` is raised with a
    detailed description of why the function is not callable.

    Note that all types must be concrete/complete.

    Examples:
        >>> bool_type = ts.ScalarType(kind=ts.ScalarKind.BOOL)
        >>> func_type = ts.FunctionType(
        ...     pos_only_args=[bool_type],
        ...     pos_or_kw_args={"foo": bool_type},
        ...     kw_only_args={},
        ...     returns=ts.VoidType(),
        ... )
        >>> accepts_args(func_type, with_args=[bool_type], with_kwargs={"foo": bool_type})
        True
        >>> accepts_args(func_type, with_args=[], with_kwargs={})
        False
    """
    if not isinstance(callable_type, ts.CallableType):
        if raise_exception:
            raise ValueError(f"Expected a callable type, got '{callable_type}'.")
        return False

    errors = function_signature_incompatibilities(callable_type, with_args, with_kwargs)
    if raise_exception:
        if len(error_list := [*errors]) > 0:
            raise ValueError(
                f"Invalid call to function of type '{callable_type}':\n"
                + ("\n".join([f"  - {error}" for error in error_list]))
            )
        return True

    return next(errors, None) is None


def needs_value_extraction(
    type_spec: ts.TypeSpec,
) -> xtyping.TypeIs[ts.NamedCollectionType | ts.TupleType]:
    return isinstance(type_spec, ts.NamedCollectionType) or (
        isinstance(type_spec, ts.TupleType)
        and any(needs_value_extraction(t) for t in type_spec.types)
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
import copy
import dataclasses


@dataclasses.dataclass(kw_only=True)
class FixMissingLocations(ast.NodeTransformer):
    """
    AST pass adding source information to every node.

    While most nodes of the Python AST have source information (lineno,
    col_offset, end_lineno, end_col_offset) after parsing, some nodes, e.g.
    :class:`ast.Pow`, do not. This pass adds this information, taking it from
    the parent node.

    Note that :func:`ast.fix_missing_locations` only adds source information to
    some ast nodes and is hence not a replacement for this pass.
    """

    _parent_nodes: list[ast.AST]

    @classmethod
    def apply(cls, node: ast.AST) -> ast.AST:
        return cls(_parent_nodes=[]).visit(node)

    def generic_visit(self, node: ast.AST) -> ast.AST:
        if not hasattr(node, "lineno"):
            node = copy.copy(node)
            parent_node = self._parent_nodes[-1]

            node.lineno = parent_node.lineno  # type: ignore[attr-defined] # we are adding the attribute which breaks type checking
            node.col_offset = parent_node.col_offset  # type: ignore[attr-defined]

            # the end positions are optional according to
            #  https://docs.python.org/3/library/ast.html#ast.AST.end_col_offset
            if hasattr(parent_node, "end_lineno"):
                node.end_lineno = parent_node.end_lineno  # type: ignore[attr-defined]
            if hasattr(parent_node, "end_col_offset"):
                node.end_col_offset = parent_node.end_col_offset  # type: ignore[attr-defined]

        self._parent_nodes.append(node)
        result = super().generic_visit(node)
        self._parent_nodes.pop()

        return result



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Generic, Literal, Optional, TypeVar, Union

import gt4py.eve as eve
from gt4py.eve import Coerced, Node, SourceLocation, SymbolName, SymbolRef, datamodels
from gt4py.eve.traits import SymbolTableTrait
from gt4py.next.ffront import dialect_ast_enums, type_specifications as ts_ffront
from gt4py.next.type_system import type_specifications as ts


class LocatedNode(Node):
    location: SourceLocation


SymbolT = TypeVar("SymbolT", bound=ts.TypeSpec)


class Symbol(eve.GenericNode, LocatedNode, Generic[SymbolT]):
    id: Coerced[SymbolName]
    type: Union[SymbolT, ts.DeferredType]  # A003
    namespace: dialect_ast_enums.Namespace = dialect_ast_enums.Namespace(
        dialect_ast_enums.Namespace.LOCAL
    )


DataTypeT = TypeVar("DataTypeT", bound=ts.DataType)
DataSymbol = Symbol[DataTypeT]

FieldTypeT = TypeVar("FieldTypeT", bound=ts.FieldType)
FieldSymbol = Symbol[FieldTypeT]

ScalarTypeT = TypeVar("ScalarTypeT", bound=ts.ScalarType)
ScalarSymbol = Symbol[ScalarTypeT]

TupleTypeT = TypeVar("TupleTypeT", bound=ts.TupleType)
TupleSymbol = Symbol[TupleTypeT]


class Expr(LocatedNode):
    type: Optional[ts.TypeSpec] = None  # A003


class BinOp(Expr):
    op: dialect_ast_enums.BinaryOperator
    left: Expr
    right: Expr


class Name(Expr):
    id: Coerced[SymbolRef]


class Call(Expr):
    func: Name
    args: list[Expr]
    kwargs: dict[str, Expr]


class Subscript(Expr):
    value: Name
    slice_: Expr


class TupleExpr(Expr):
    elts: list[Expr]


class Attribute(Expr):
    attr: str
    value: Expr


class Constant(Expr):
    value: Any  # TODO(tehrengruber): be more restrictive


class Dict(Expr):
    keys_: list[Union[Name | Attribute]]
    values_: list[TupleExpr]

    @datamodels.root_validator
    @classmethod
    def keys_values_length_validation(cls: type["Dict"], instance: "Dict") -> None:
        if len(instance.keys_) != len(instance.values_):
            raise ValueError("`Dict` must have same number of keys as values.")


class Slice(Expr):
    lower: Optional[Constant]
    upper: Optional[Constant]
    step: Literal[None]


class Stmt(LocatedNode): ...


class Program(LocatedNode, SymbolTableTrait):
    id: Coerced[SymbolName]
    type: Union[ts_ffront.ProgramType, ts.DeferredType]  # A003
    params: list[DataSymbol]
    body: list[Call]
    closure_vars: list[Symbol]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import collections
import dataclasses
import enum
import functools
import operator
from typing import Optional

from gt4py import eve
from gt4py.next import common, utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
    misc as ir_misc,
)
from gt4py.next.iterator.transforms import (
    cse,
    fixed_point_transformation,
    inline_center_deref_lift_vars,
    inline_lambdas,
    inline_lifts,
    merge_let,
    trace_shifts,
)
from gt4py.next.iterator.type_system import inference as type_inference
from gt4py.next.type_system import type_info, type_specifications as ts


def _merge_arguments(
    args1: dict[str, itir.Expr], args2: dict[str, itir.Expr]
) -> dict[str, itir.Expr]:
    new_args = {**args1}
    for stencil_param, stencil_arg in args2.items():
        assert stencil_param not in new_args
        new_args[stencil_param] = stencil_arg
    return new_args


def _is_tuple_expr_of_literals(expr: itir.Expr):
    if cpm.is_call_to(expr, "make_tuple"):
        return all(_is_tuple_expr_of_literals(arg) for arg in expr.args)
    if cpm.is_call_to(expr, "tuple_get"):
        return _is_tuple_expr_of_literals(expr.args[1])
    return isinstance(expr, itir.Literal)


def _inline_as_fieldop_arg(
    arg: itir.Expr, *, uids: utils.IDGeneratorPool
) -> tuple[itir.Expr, dict[str, itir.Expr]]:
    assert cpm.is_applied_as_fieldop(arg)
    arg = ir_misc.canonicalize_as_fieldop(arg)

    stencil, *_ = arg.fun.args  # type: ignore[attr-defined]  # ensured by `is_applied_as_fieldop`
    inner_args: list[itir.Expr] = arg.args
    extracted_args: dict[str, itir.Expr] = {}  # mapping from outer-stencil param to arg

    stencil_params: list[itir.Sym] = []
    stencil_body: itir.Expr = stencil.expr

    for inner_param, inner_arg in zip(stencil.params, inner_args, strict=True):
        if isinstance(inner_arg, itir.Literal):
            # note: only literals, not all scalar expressions are required as it doesn't make sense
            # for them to be computed per grid point.
            stencil_body = im.let(inner_param, im.promote_to_const_iterator(inner_arg))(
                stencil_body
            )
        else:
            stencil_params.append(inner_param)
            new_outer_stencil_param = next(uids["__iasfop"])
            extracted_args[new_outer_stencil_param] = inner_arg

    return im.lift(im.lambda_(*stencil_params)(stencil_body))(
        *extracted_args.keys()
    ), extracted_args


def _deduplicate_as_fieldop_args(
    args: dict[str, itir.Expr], stencil_body: itir.Expr
) -> tuple[dict[str, itir.Expr], itir.Expr]:
    new_args_inverted: dict[itir.Expr, list[str]] = collections.defaultdict(list)
    for name, arg in args.items():
        new_args_inverted[arg].append(name)

    new_args: dict[str, itir.Expr] = {}
    new_stencil_body = stencil_body
    for arg, names in new_args_inverted.items():
        # put internal names at the end
        new_name, *aliases = sorted(names, key=lambda s: (s.startswith("_"), s))
        new_args[new_name] = arg
        if aliases:
            new_stencil_body = im.let(*((alias, new_name) for alias in aliases))(new_stencil_body)

    return new_args, new_stencil_body


def _prettify_as_fieldop_args(
    args: dict[str, itir.Expr], stencil_body: itir.Expr
) -> tuple[dict[str, itir.Expr], itir.Expr]:
    new_args: dict[str, itir.Expr] = {}
    remap_table: dict[str, str] = {}
    for name, arg in args.items():
        if isinstance(arg, itir.SymRef):
            assert arg.id not in new_args  # ensured by deduplication
            new_args[arg.id] = arg
            remap_table[name] = arg.id
        else:
            new_args[name] = arg

    return new_args, im.let(*remap_table.items())(stencil_body)


def fuse_as_fieldop(
    expr: itir.Expr,
    eligible_args: list[bool],
    *,
    offset_provider_type: common.OffsetProviderType,
    enable_cse: bool,
    uids: utils.IDGeneratorPool,
) -> itir.Expr:
    assert cpm.is_applied_as_fieldop(expr)

    assert isinstance(expr.fun.args[0], itir.Lambda) or cpm.is_call_to(expr.fun.args[0], "scan")
    stencil, restore_scan = ir_misc.unwrap_scan(expr.fun.args[0])

    domain = expr.fun.args[1] if len(expr.fun.args) > 1 else None

    args: list[itir.Expr] = expr.args

    new_args: dict[str, itir.Expr] = {}
    new_stencil_body: itir.Expr = stencil.expr

    for eligible, stencil_param, arg in zip(eligible_args, stencil.params, args, strict=True):
        if eligible:
            if cpm.is_applied_as_fieldop(arg):
                pass
            elif cpm.is_call_to(arg, "if_"):
                # transform scalar `if` into per-grid-point `if`
                # TODO(tehrengruber): revisit if we want to inline if_
                arg = im.op_as_fieldop("if_")(*arg.args)
            elif _is_tuple_expr_of_literals(arg):
                arg = im.op_as_fieldop(im.lambda_()(arg))()
            else:
                raise NotImplementedError()

            inline_expr, extracted_args = _inline_as_fieldop_arg(arg, uids=uids)

            new_stencil_body = im.let(stencil_param, inline_expr)(new_stencil_body)

            new_args = _merge_arguments(new_args, extracted_args)
        else:
            # just a safety check if typing information is available
            type_inference.reinfer(arg)
            if arg.type and not isinstance(arg.type, ts.DeferredType):
                assert isinstance(arg.type, ts.TypeSpec)

                assert not isinstance(
                    type_info.apply_to_primitive_constituents(type_info.extract_dtype, arg.type),
                    ts.ListType,
                )
            new_args = _merge_arguments(new_args, {stencil_param.id: arg})

    new_args, new_stencil_body = _deduplicate_as_fieldop_args(new_args, new_stencil_body)
    new_args, new_stencil_body = _prettify_as_fieldop_args(new_args, new_stencil_body)

    stencil = im.lambda_(*new_args.keys())(new_stencil_body)
    new_stencil = restore_scan(stencil)

    # simplify stencil directly to keep the tree small
    new_stencil = inline_lambdas.InlineLambdas.apply(
        new_stencil, opcount_preserving=True, force_inline_lift_args=False
    )
    new_stencil = inline_center_deref_lift_vars.InlineCenterDerefLiftVars.apply(
        new_stencil, is_stencil=True, uids=uids
    )  # to keep the tree small
    new_stencil = merge_let.MergeLet().visit(new_stencil)
    new_stencil = inline_lambdas.InlineLambdas.apply(
        new_stencil, opcount_preserving=True, force_inline_lift_args=True
    )
    new_stencil = inline_lifts.InlineLifts().visit(new_stencil)

    new_node = im.as_fieldop(new_stencil, domain)(*new_args.values())
    if enable_cse:
        # TODO(havogt): We should investigate how to keep the tree small without having to run CSE.
        new_node = cse.CommonSubexpressionElimination.apply(
            new_node, within_stencil=False, uids=uids, offset_provider_type=offset_provider_type
        )

    return new_node


def _arg_inline_predicate(node: itir.Expr, shifts: set[tuple[itir.OffsetLiteral, ...]]) -> bool:
    if _is_tuple_expr_of_literals(node):
        return True

    if (
        is_applied_fieldop := cpm.is_applied_as_fieldop(node)
        and not cpm.is_call_to(node.fun.args[0], "scan")
    ) or cpm.is_call_to(node, "if_"):
        # always inline arg if it is an applied fieldop with only a single arg
        if is_applied_fieldop and len(node.args) == 1:
            return True
        # argument is never used, will be removed when inlined
        if len(shifts) == 0:
            return True
        # applied fieldop with list return type must always be inlined as no backend supports this
        type_inference.reinfer(node)
        assert isinstance(node.type, ts.TypeSpec)
        dtype = type_info.apply_to_primitive_constituents(type_info.extract_dtype, node.type)
        if isinstance(dtype, ts.ListType):
            return True
        # only accessed at the center location
        if shifts in [set(), {()}]:
            return True
        # TODO(tehrengruber): Disabled as the InlineCenterDerefLiftVars does not support this yet
        #  and it would increase the size of the tree otherwise.
        # if len(shifts) == 1 and not any(
        #     trace_shifts.Sentinel.ALL_NEIGHBORS in access for access in shifts
        # ):
        #     return True  # noqa: ERA001 [commented-out-code]

    return False


def _make_tuple_element_inline_predicate(node: itir.Expr):
    if cpm.is_applied_as_fieldop(node):  # field, or tuple of fields
        return True
    if isinstance(node.type, ts.FieldType) and isinstance(node, itir.SymRef):
        return True
    return False


@dataclasses.dataclass(frozen=True, kw_only=True)
class FuseAsFieldOp(
    fixed_point_transformation.CombinedFixedPointTransform, eve.PreserveLocationVisitor
):
    """
    Merge multiple `as_fieldop` calls into one.

    >>> from gt4py import next as gtx
    >>> from gt4py.next import utils
    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> IDim = gtx.Dimension("IDim")
    >>> field_type = ts.FieldType(dims=[IDim], dtype=ts.ScalarType(kind=ts.ScalarKind.INT32))
    >>> d = im.domain("cartesian_domain", {IDim: (0, 1)})
    >>> nested_as_fieldop = im.op_as_fieldop("plus", d)(
    ...     im.op_as_fieldop("multiplies", d)(
    ...         im.ref("inp1", field_type), im.ref("inp2", field_type)
    ...     ),
    ...     im.ref("inp3", field_type),
    ... )
    >>> print(nested_as_fieldop)
    as_fieldop(λ(__arg0, __arg1) → ·__arg0 + ·__arg1, c⟨ IDimₕ: [0, 1[ ⟩)(
      as_fieldop(λ(__arg0, __arg1) → ·__arg0 × ·__arg1, c⟨ IDimₕ: [0, 1[ ⟩)(inp1, inp2), inp3
    )
    >>> print(
    ...     FuseAsFieldOp.apply(
    ...         nested_as_fieldop,
    ...         offset_provider_type={},
    ...         allow_undeclared_symbols=True,
    ...         uids=utils.IDGeneratorPool(),
    ...     )
    ... )
    as_fieldop(λ(inp1, inp2, inp3) → ·inp1 × ·inp2 + ·inp3, c⟨ IDimₕ: [0, 1[ ⟩)(inp1, inp2, inp3)
    """  # noqa: RUF002  # ignore ambiguous multiplication character

    class Transformation(enum.Flag):
        #: Let `f_expr` be an expression with list dtype then
        #: `let(f, f_expr) -> as_fieldop(...)(f)` -> `as_fieldop(...)(f_expr)`
        FUSE_MAKE_TUPLE = enum.auto()
        #: `as_fieldop(...)(as_fieldop(...)(a, b), c)`
        #: -> as_fieldop(fused_stencil)(a, b, c)
        FUSE_AS_FIELDOP = enum.auto()
        INLINE_LET_VARS_OPCOUNT_PRESERVING = enum.auto()

        @classmethod
        def all(self) -> FuseAsFieldOp.Transformation:
            return functools.reduce(operator.or_, self.__members__.values())

    PRESERVED_ANNEX_ATTRS = ("domain",)
    REINFER_TYPES = True

    enabled_transformations = Transformation.all()

    uids: utils.IDGeneratorPool
    offset_provider_type: common.OffsetProviderType
    enable_cse: bool  # option to disable is mainly for testing purposes

    @classmethod
    def apply(
        cls,
        node: itir.Program,
        *,
        offset_provider_type: common.OffsetProviderType,
        uids: utils.IDGeneratorPool,
        allow_undeclared_symbols=False,
        within_set_at_expr: Optional[bool] = None,
        enabled_transformations: Optional[Transformation] = None,
        enable_cse: bool = True,
    ):
        enabled_transformations = enabled_transformations or cls.enabled_transformations

        node = type_inference.infer(
            node,
            offset_provider_type=offset_provider_type,
            allow_undeclared_symbols=allow_undeclared_symbols,
        )

        if within_set_at_expr is None:
            within_set_at_expr = not isinstance(node, itir.Program)

        new_node = cls(
            uids=uids,
            enabled_transformations=enabled_transformations,
            offset_provider_type=offset_provider_type,
            enable_cse=enable_cse,
        ).visit(node, within_set_at_expr=within_set_at_expr)
        # The `FuseAsFieldOp` pass does not fully preserve the type information yet. In particular
        # for the generated lifts this is tricky and error-prone. For simplicity, we just reinfer
        # everything here ensuring later passes can use the information.
        new_node = type_inference.infer(
            new_node,
            offset_provider_type=offset_provider_type,
            allow_undeclared_symbols=allow_undeclared_symbols,
        )
        return new_node

    def transform_fuse_make_tuple(self, node: itir.Node, **kwargs):
        if not cpm.is_call_to(node, "make_tuple"):
            return None

        for arg in node.args:
            type_inference.reinfer(arg)
            assert not isinstance(arg.type, ts.FieldType) or (
                hasattr(arg.annex, "domain")
                and isinstance(arg.annex.domain, domain_utils.SymbolicDomain)
            )

        eligible_els = [_make_tuple_element_inline_predicate(arg) for arg in node.args]
        field_args = [arg for i, arg in enumerate(node.args) if eligible_els[i]]
        distinct_domains = set(arg.annex.domain.as_expr() for arg in field_args)
        if len(distinct_domains) != len(field_args):
            new_els: list[itir.Expr | None] = [None for _ in node.args]
            field_args_by_domain: dict[itir.FunCall, list[tuple[int, itir.Expr]]] = {}
            for i, arg in enumerate(node.args):
                if eligible_els[i]:
                    assert isinstance(arg.annex.domain, domain_utils.SymbolicDomain)
                    domain = arg.annex.domain.as_expr()
                    field_args_by_domain.setdefault(domain, [])
                    field_args_by_domain[domain].append((i, arg))
                else:
                    new_els[i] = arg  # keep as is

            if len(field_args_by_domain) == 1 and all(eligible_els):
                # if we only have a single domain covering all args we don't need to create an
                # unnecessary let
                ((domain, inner_field_args),) = field_args_by_domain.items()
                new_node = im.op_as_fieldop(lambda *args: im.make_tuple(*args), domain)(
                    *(arg for _, arg in inner_field_args)
                )
                new_node = self.visit(new_node, **{**kwargs, "recurse": False})
            else:
                let_vars = {}
                for domain, inner_field_args in field_args_by_domain.items():
                    if len(inner_field_args) > 1:
                        var = next(self.uids["__fasfop"])
                        fused_args = im.op_as_fieldop(lambda *args: im.make_tuple(*args), domain)(
                            *(arg for _, arg in inner_field_args)
                        )
                        type_inference.reinfer(arg)
                        # don't recurse into nested args, but only consider newly created `as_fieldop`
                        # note: this will always inline (as we inline center accessed)
                        let_vars[var] = self.visit(fused_args, **{**kwargs, "recurse": False})
                        for outer_tuple_idx, (inner_tuple_idx, _) in enumerate(inner_field_args):
                            new_el = im.tuple_get(outer_tuple_idx, var)
                            new_el.annex.domain = domain_utils.SymbolicDomain.from_expr(domain)
                            new_els[inner_tuple_idx] = new_el
                    else:
                        i, arg = inner_field_args[0]
                        new_els[i] = arg
                assert not any(el is None for el in new_els)
                assert let_vars
                new_node = im.let(*let_vars.items())(im.make_tuple(*new_els))
                new_node = inline_lambdas.inline_lambda(new_node, opcount_preserving=True)
            return new_node
        return None

    def transform_fuse_as_fieldop(self, node: itir.Node, **kwargs):
        if cpm.is_applied_as_fieldop(node):
            node = ir_misc.canonicalize_as_fieldop(node)
            stencil = node.fun.args[0]  # type: ignore[attr-defined]  # ensure cpm.is_applied_as_fieldop
            assert isinstance(stencil, itir.Lambda) or cpm.is_call_to(stencil, "scan")
            args: list[itir.Expr] = node.args
            shifts = trace_shifts.trace_stencil(stencil, num_args=len(args))

            eligible_els = [
                _arg_inline_predicate(arg, arg_shifts)
                for arg, arg_shifts in zip(args, shifts, strict=True)
            ]
            if any(eligible_els):
                return self.visit(
                    fuse_as_fieldop(
                        node,
                        eligible_els,
                        uids=self.uids,
                        offset_provider_type=self.offset_provider_type,
                        enable_cse=self.enable_cse,
                    ),
                    **{**kwargs, "recurse": False},
                )
        return None

    def transform_inline_let_vars_opcount_preserving(self, node: itir.Node, **kwargs):
        # when multiple `as_fieldop` calls are fused that use the same argument, this argument
        # might become referenced once only. In order to be able to continue fusing such arguments
        # try inlining here.
        if cpm.is_let(node):
            new_node = inline_lambdas.inline_lambda(node, opcount_preserving=True)
            if new_node is not node:  # nothing has been inlined
                return self.visit(new_node, **kwargs)

        return None

    def generic_visit(self, node, **kwargs):
        if cpm.is_applied_as_fieldop(node):  # don't descend in stencil
            return im.as_fieldop(*node.fun.args)(*self.visit(node.args, **kwargs))

        # TODO(tehrengruber): This is a common pattern that should be absorbed in
        #  `FixedPointTransformation`.
        if kwargs.get("recurse", True):
            return super().generic_visit(node, **kwargs)
        else:
            return node

    def visit(self, node, **kwargs):
        if isinstance(node, itir.SetAt):
            return itir.SetAt(
                expr=self.visit(node.expr, **kwargs | {"within_set_at_expr": True}),
                # rest doesn't need to be visited
                domain=node.domain,
                target=node.target,
            )

        # don't execute transformations unless inside `SetAt` node
        if not kwargs.get("within_set_at_expr"):
            return self.generic_visit(node, **kwargs)

        # inline all fields with list dtype. This needs to happen before the children are visited
        # such that the `as_fieldop` can be fused.
        # TODO(tehrengruber): what should we do in case the field with list dtype is a let itself?
        #  This could duplicate other expressions which we did not intend to duplicate.
        # TODO(tehrengruber): This should be moved into a `transform_` method, but
        #  `FixedPointTransformation` does not support pre-order transformations yet.
        if cpm.is_let(node):
            for arg in node.args:
                type_inference.reinfer(arg)
            eligible_els = [
                isinstance(arg.type, ts.FieldType) and isinstance(arg.type.dtype, ts.ListType)
                for arg in node.args
            ]
            if any(eligible_els):
                node = inline_lambdas.inline_lambda(node, eligible_params=eligible_els)
                return self.visit(node, **kwargs)

        node = super().visit(node, **kwargs)

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Iterable
from typing import Callable, Optional, TypeVar

from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.type_system import type_specifications as ts


# TODO(tehrengruber): The code quality of this function is poor. We should rewrite it.
def process_elements(
    process_func: Callable[..., itir.Expr],
    objs: itir.Expr | Iterable[itir.Expr],
    current_el_type: ts.TypeSpec,
    arg_types: Optional[Iterable[ts.TypeSpec]] = None,
) -> itir.FunCall:
    """
    Recursively applies a processing function to all primitive constituents of a tuple or
    named collection.

    Arguments:
        process_func: A callable that takes an itir.Expr representing a leaf-element of `objs`.
            If multiple `objs` are given the callable takes equally many arguments.
        objs: The object whose elements are to be transformed.
        current_el_type: A type with the same structure as the elements of `objs`. The leaf-types
            are not used and thus not relevant.
        arg_types: If provided, a tuple of the type of each argument is passed to `process_func` as last argument.
            Note, that `arg_types` might coincide with `(current_el_type,)*len(objs)`, but not necessarily,
            in case of implicit broadcasts.
    """
    if isinstance(objs, itir.Expr):
        objs = (objs,)

    var_names = tuple(f"__val_{obj.fingerprint()}" for obj in objs)
    # Note: The same `var_name` might appear multiple times if the same object appears multiple
    # times. Since we use a dict to collect the bound variables, this is fine.
    bound_vars = {var_name: obj for var_name, obj in zip(var_names, objs)}
    body = _process_elements_impl(
        process_func,
        tuple(im.ref(var_name) for var_name in var_names),
        current_el_type,
        arg_types=arg_types,
    )

    return im.let(*bound_vars.items())(body)


T = TypeVar("T", bound=itir.Expr, covariant=True)


def _process_elements_impl(
    process_func: Callable[..., itir.Expr],
    _current_el_exprs: Iterable[T],
    current_el_type: ts.TypeSpec,
    arg_types: Optional[Iterable[ts.TypeSpec]],
) -> itir.Expr:
    if isinstance(current_el_type, (ts.TupleType, ts.NamedCollectionType)):
        result = im.make_tuple(
            *(
                _process_elements_impl(
                    process_func,
                    tuple(
                        im.tuple_get(i, current_el_expr) for current_el_expr in _current_el_exprs
                    ),
                    current_el_type.types[i],
                    arg_types=tuple(arg_t.types[i] for arg_t in arg_types)  # type: ignore[attr-defined] # guaranteed by the requirement that `current_el_type` and each element of `arg_types` have the same tuple structure
                    if arg_types is not None
                    else None,
                )
                for i in range(len(current_el_type.types))
            )
        )
    else:
        if arg_types is not None:
            result = process_func(*_current_el_exprs, arg_types)
        else:
            result = process_func(*_current_el_exprs)

    return result



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from .push_vertical_map_down import PushVerticalMapDown
from .swap_horizontal_maps import SwapHorizontalMaps


__all__ = ["PushVerticalMapDown", "SwapHorizontalMaps"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Any, Optional

import dace
import factory

from gt4py._core import definitions as core_defs
from gt4py.next import common, config
from gt4py.next.instrumentation import metrics
from gt4py.next.iterator import ir as itir, transforms as itir_transforms
from gt4py.next.otf import code_specs, definitions, stages, workflow
from gt4py.next.otf.binding import interface
from gt4py.next.program_processors.runners.dace import (
    lowering as gtx_dace_lowering,
    sdfg_args as gtx_dace_args,
    transformations as gtx_transformations,
)
from gt4py.next.program_processors.runners.dace.workflow import common as gtx_wfdcommon
from gt4py.next.type_system import type_specifications as ts


def find_constant_symbols(
    ir: itir.Program,
    sdfg: dace.SDFG,
    offset_provider_type: common.OffsetProviderType,
    disable_field_origin_on_program_arguments: bool,
    unstructured_horizontal_has_unit_stride: bool,
) -> dict[str, int]:
    """Helper function to find symbols to replace with constant values."""
    constant_symbols: dict[str, int] = {}

    if unstructured_horizontal_has_unit_stride:
        # Search the stride symbols corresponding to the horizontal dimension
        for p in ir.params:
            if isinstance(p.type, ts.FieldType):
                h_dims = [dim for dim in p.type.dims if dim.kind == common.DimensionKind.HORIZONTAL]
                if len(h_dims) == 0:
                    continue
                elif len(h_dims) == 1:
                    dim = h_dims[0]
                else:
                    raise NotImplementedError(
                        f"Unsupported field with multiple horizontal dimensions '{p}'."
                    )
                sdfg_stride_symbol = gtx_dace_args.field_stride_symbol(str(p.id), dim)
                constant_symbols[sdfg_stride_symbol.name] = 1
        # Same for connectivity tables, for which the first dimension is always horizontal
        for offset, conn_type in offset_provider_type.items():
            if (
                isinstance(conn_type, common.NeighborConnectivityType)
                and (conn_id := gtx_dace_args.connectivity_identifier(offset)) in sdfg.arrays
            ):
                assert not sdfg.arrays[conn_id].transient
                assert conn_type.source_dim.kind == common.DimensionKind.HORIZONTAL
                sdfg_stride_symbol = gtx_dace_args.field_stride_symbol(
                    conn_id, conn_type.source_dim, offset_provider_type
                )
                constant_symbols[sdfg_stride_symbol.name] = 1

    if disable_field_origin_on_program_arguments:
        # collect symbols used as range start for all program arguments
        for p in ir.params:
            if isinstance(p.type, ts.TupleType):
                psymbols = [
                    sym
                    for sym in gtx_dace_lowering.flatten_tuple_fields(p.id, p.type)
                    if isinstance(sym.type, ts.FieldType)
                ]
            elif isinstance(p.type, ts.FieldType):
                psymbols = [p]
            else:
                psymbols = []
            for psymbol in psymbols:
                assert isinstance(psymbol.type, ts.FieldType)
                if len(psymbol.type.dims) == 0:
                    # zero-dimensional field
                    continue
                # set all range start symbols to constant value 0
                sdfg_origin_symbols = [
                    gtx_dace_args.range_start_symbol(str(psymbol.id), dim)
                    for dim in psymbol.type.dims
                ]
                constant_symbols |= {sdfg_symbol.name: 0 for sdfg_symbol in sdfg_origin_symbols}

    return constant_symbols


def make_sdfg_call_async(sdfg: dace.SDFG, gpu: bool) -> None:
    """Configure an SDFG to immediately return once all work has been scheduled.

    This means that `CompiledSDFG.fast_call()` will return immediately after all
    computations have been _scheduled_ on the device. This function only has an effect
    for work that runs on the GPU. Furthermore, all work is scheduled on the
    default stream.

    Todo: Revisit this function once DaCe changes its behaviour in this regard.
    """

    # This is only a problem on GPU.
    # TODO(phimuell): Figuring out what about OpenMP.
    if not gpu:
        return

    assert dace.Config.get("compiler.cuda.max_concurrent_streams") == -1, (
        f"Expected `max_concurrent_streams == -1` but it was `{dace.Config.get('compiler.cuda.max_concurrent_streams')}`."
    )

    # NOTE: We are using the default stream this means that _**currently**_ the launch is
    #   already asynchronous, see [DaCe issue#2120](https://github.com/spcl/dace/issues/2120)
    #   for more. However, DaCe still [generates streams internally](https://github.com/spcl/dace/blob/54c935cfe74a52c5107dc91680e6201ddbf86821/dace/codegen/targets/cuda.py#L467).
    #   Thus to be absolutely sure we will no set all streams, DaCe uses to the default
    #   stream.
    # NOTE: Another important note here is, that it looks as if no synchronization
    #   what soever between states is generated if the default stream is used. This
    #   might lead to problems if a GPU kernel computes something that is needed for a
    #   condition of an interstate edge. However, we should not have that case.
    # TODO(phimuell, edopao): Make sure if this is really the case.
    dace_gpu_backend = dace.Config.get("compiler.cuda.backend")
    assert dace_gpu_backend in ["cuda", "hip"], f"GPU backend '{dace_gpu_backend}' is unknown."
    sdfg.append_init_code(
        f"__dace_gpu_set_all_streams(__state, {dace_gpu_backend}StreamDefault);",
        location="cuda",
    )


def _has_gpu_schedule(sdfg: dace.SDFG) -> bool:
    """Check if any node (e.g. maps) of the given SDFG is scheduled on GPU."""
    return any(
        getattr(node, "schedule", dace.dtypes.ScheduleType.Default) in dace.dtypes.GPU_SCHEDULES
        for node, _ in sdfg.all_nodes_recursive()
    )


def _make_if_region_for_metrics_collection(
    name: str,
    metrics_level: str,
    sdfg: dace.SDFG,
) -> tuple[dace.state.ConditionalBlock, dace.state.SDFGState]:
    """
    Helper function to create a conditional block in the given SDFG, with only one
    branch to be executed if 'metric_level >= metrics.PERFORMANCE' is true.
    """
    if_region = dace.sdfg.state.ConditionalBlock(name)
    sdfg.add_node(if_region, ensure_unique_name=True)
    then_body = dace.sdfg.state.ControlFlowRegion(f"{if_region.label}_collect_metrics", sdfg=sdfg)
    then_state = then_body.add_state(f"{if_region.label}_collect_metrics")
    if_region.add_branch(
        dace.sdfg.state.CodeBlock(f"{metrics_level} >= {metrics.PERFORMANCE}"), then_body
    )
    return if_region, then_state


def add_instrumentation(sdfg: dace.SDFG, gpu: bool) -> None:
    """
    Instrument SDFG with measurement of total execution time.

    We measure the execution time of one GT4Py program by instrumenting the top-level
    SDFG with a cpp timer (std::chrono). This timer measures only the computation
    time, it does not include the overhead of calling the SDFG from Python.

    The execution time is measured in seconds and represented as a 'float64' value.
    It is written to the global array 'SDFG_ARG_METRIC_COMPUTE_TIME'.
    """
    output, _ = sdfg.add_array(gtx_wfdcommon.SDFG_ARG_METRIC_COMPUTE_TIME, [1], dace.float64)
    start_time, _ = sdfg.add_scalar("gt_start_time", dace.int64, transient=True)
    metrics_level = sdfg.add_symbol(gtx_wfdcommon.SDFG_ARG_METRIC_LEVEL, dace.int32)

    #### 1. Synchronize the CUDA device, in order to wait for kernels completion.
    # Even when the target device is GPU, it can happen that dace emits code without
    # GPU kernels. In this case, the cuda headers are not imported and the SDFG is
    # compiled as plain C++. Therefore, we also check here the schedule of SDFG maps.
    if gpu and _has_gpu_schedule(sdfg):
        dace_gpu_backend = dace.Config.get("compiler.cuda.backend")
        assert dace_gpu_backend in ["cuda", "hip"], f"GPU backend '{dace_gpu_backend}' is unknown."

        # NOTE: We should actually wrap the `DeviceSynchronize` function inside a
        #   `DACE_GPU_CHECK()` macro. However, this only works in GPU context, but
        #   here we are in CPU context. Thus we cannot do it.
        sync_code = f"{dace_gpu_backend}DeviceSynchronize();"
        has_side_effects = True

    else:
        sync_code = "/* The SDFG execution should already be synchronized */"
        has_side_effects = False

    #### 2. Timestamp the SDFG entry point.
    start_block = sdfg.start_block
    entry_if_region, begin_state = _make_if_region_for_metrics_collection(
        "metrics_entry", metrics_level, sdfg
    )
    sdfg.add_edge(entry_if_region, start_block, dace.InterstateEdge())
    sdfg.start_block = sdfg.node_id(entry_if_region)
    assert sdfg.start_block is entry_if_region

    tlet_start_timer = begin_state.add_tasklet(
        "gt_start_timer",
        inputs={},
        outputs={"time"},
        code=f"""\
{sync_code}
auto now = std::chrono::high_resolution_clock::now();
time = std::chrono::duration_cast<std::chrono::nanoseconds>(
        now.time_since_epoch()
    ).count();
        """,
        language=dace.dtypes.Language.CPP,
        side_effects=has_side_effects,
    )
    begin_state.add_edge(
        tlet_start_timer,
        "time",
        begin_state.add_access(start_time),
        None,
        dace.Memlet(f"{start_time}[0]"),
    )

    #### 3. Collect the SDFG end timestamp and produce the compute metric.
    exit_if_region, end_state = _make_if_region_for_metrics_collection(
        "metrics_exit", metrics_level, sdfg
    )
    for sink_node in sdfg.sink_nodes():
        if sink_node is exit_if_region:
            continue
        sdfg.add_edge(sink_node, exit_if_region, dace.InterstateEdge())
    assert sdfg.in_degree(exit_if_region) > 0

    # Populate the branch that computes the stencil time metric
    tlet_stop_timer = end_state.add_tasklet(
        "gt_stop_timer",
        inputs={"run_cpp_start_time"},
        outputs={"duration"},
        code=f"""\
{sync_code}
auto now = std::chrono::high_resolution_clock::now();
auto run_cpp_end_time = std::chrono::duration_cast<std::chrono::nanoseconds>(
        now.time_since_epoch()
    ).count();
duration = static_cast<double>(run_cpp_end_time - run_cpp_start_time) * 1.e-9;
        """,
        language=dace.dtypes.Language.CPP,
        side_effects=has_side_effects,
    )
    end_state.add_edge(
        end_state.add_access(start_time),
        None,
        tlet_stop_timer,
        "run_cpp_start_time",
        dace.Memlet(f"{start_time}[0]"),
    )
    end_state.add_edge(
        tlet_stop_timer,
        "duration",
        end_state.add_access(output),
        None,
        dace.Memlet(f"{output}[0]"),
    )

    if gpu and _has_gpu_schedule(sdfg) and config.ADD_GPU_TRACE_MARKERS:
        sdfg.instrument = dace.dtypes.InstrumentationType.GPU_TX_MARKERS
        for node, _ in sdfg.all_nodes_recursive():
            if isinstance(
                node, dace.nodes.MapEntry
            ):  # Add ranges to scopes and maps that are NOT scheduled to the GPU
                node.instrument = dace.dtypes.InstrumentationType.GPU_TX_MARKERS
            elif isinstance(node, dace.sdfg.state.SDFGState):
                node.instrument = dace.dtypes.InstrumentationType.GPU_TX_MARKERS

    # Check SDFG validity after applying the above changes.
    # Normally, we do not call `SDFGState.add_tasklet()` directly, instead we call
    #  the wrapper provided by `DataflowBuilder`, that modifies the tasklet connectors
    #  to avoid name conflicts with program symbols. However, this method is not
    #  available here, so we have to call the underlying DaCe function directly.
    #  We now run `validate()` to make sure that no name conflict was introduced.
    sdfg.validate()


def make_sdfg_call_sync(sdfg: dace.SDFG, gpu: bool) -> None:
    """Process the SDFG such that the call is synchronous.

    This means that `CompiledSDFG.fast_call()` will return only after all computations
    have _finished_ and the results are available. This function only has an effect for
    work that runs on the GPU. Furthermore, all work is scheduled on the default stream.
    """

    if not gpu:
        # This is only a problem on GPU. Dace uses OpenMP on CPU and
        # the OpenMP parallel region creates a synchronization point.
        return
    elif not _has_gpu_schedule(sdfg):
        # Even when the target device is GPU, it can happen that dace
        # emits code without GPU kernels. In this case, the cuda headers
        # are not imported and the SDFG is compiled as plain C++.
        return

    assert dace.Config.get("compiler.cuda.max_concurrent_streams") == -1, (
        f"Expected `max_concurrent_streams == -1` but it was `{dace.Config.get('compiler.cuda.max_concurrent_streams')}`."
    )

    # If we are using the default stream, things are a bit simpler/harder. For some
    #  reasons when using the default stream, DaCe seems to skip _all_ synchronization,
    #  for more see [DaCe issue#2120](https://github.com/spcl/dace/issues/2120).
    #  Thus the `CompiledSDFG.fast_call()` call is truly asynchronous, i.e. just
    #  launches the kernels and then exist. Thus we have to add a synchronization
    #  at the end to have a synchronous call. We can not use `SDFG.append_exit_code()`
    #  because that code is only run at the `exit()` stage, not after a call. Thus we
    #  will generate an SDFGState that contains a Tasklet with the sync call.
    sync_state = sdfg.add_state("sync_state")
    for sink_node in sdfg.sink_nodes():
        if sink_node is sync_state:
            continue
        sdfg.add_edge(sink_node, sync_state, dace.InterstateEdge())
    assert sdfg.in_degree(sync_state) > 0

    # NOTE: Since the synchronization is done through the Tasklet explicitly,
    #   we can disable synchronization for the last state. Might be useless.
    sync_state.nosync = True

    # NOTE: We should actually wrap the `StreamSynchronize` function inside a
    #   `DACE_GPU_CHECK()` macro. However, this only works in GPU context, but
    #   here we are in CPU context. Thus we can not do it.
    dace_gpu_backend = dace.Config.get("compiler.cuda.backend")
    assert dace_gpu_backend in ["cuda", "hip"], f"GPU backend '{dace_gpu_backend}' is unknown."
    sync_state.add_tasklet(
        "sync_tlet",
        inputs=set(),
        outputs=set(),
        code=f"{dace_gpu_backend}StreamSynchronize({dace_gpu_backend}StreamDefault);",
        language=dace.dtypes.Language.CPP,
        side_effects=True,
    )

    # DaCe [still generates a stream](https://github.com/spcl/dace/blob/54c935cfe74a52c5107dc91680e6201ddbf86821/dace/codegen/targets/cuda.py#L467)
    #  despite not using it. Thus to be absolutely sure, we will not set that stream
    #  to the default stream.
    sdfg.append_init_code(
        f"__dace_gpu_set_all_streams(__state, {dace_gpu_backend}StreamDefault);",
        location="cuda",
    )


@dataclasses.dataclass(frozen=True)
class DaCeTranslator(
    workflow.ChainableWorkflowMixin[
        definitions.CompilableProgramDef,
        stages.ProgramSource[code_specs.SDFGCodeSpec],
    ],
    definitions.TranslationStep[code_specs.SDFGCodeSpec],
):
    device_type: core_defs.DeviceType
    auto_optimize: bool
    auto_optimize_args: dict[str, Any] | None
    async_sdfg_call: bool
    unstructured_horizontal_has_unit_stride: bool
    use_metrics: bool

    disable_itir_transforms: bool = False
    disable_field_origin_on_program_arguments: bool = False
    use_max_domain_range_on_unstructured_shift: bool | None = None

    def generate_sdfg(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> dace.SDFG:
        with gtx_wfdcommon.dace_context(device_type=self.device_type):
            return self._generate_sdfg_without_configuring_dace(*args, **kwargs)

    def _generate_sdfg_without_configuring_dace(
        self,
        ir: itir.Program,
        offset_provider: common.OffsetProvider,
        column_axis: Optional[common.Dimension],
    ) -> dace.SDFG:
        if not self.disable_itir_transforms:
            ir = itir_transforms.apply_fieldview_transforms(
                ir,
                use_max_domain_range_on_unstructured_shift=self.use_max_domain_range_on_unstructured_shift,
                offset_provider=offset_provider,
            )
        offset_provider_type = common.offset_provider_to_type(offset_provider)
        on_gpu = self.device_type != core_defs.DeviceType.CPU

        sdfg = gtx_dace_lowering.build_sdfg_from_gtir(ir, offset_provider_type, column_axis)

        constant_symbols = find_constant_symbols(
            ir,
            sdfg,
            offset_provider_type,
            self.disable_field_origin_on_program_arguments,
            self.unstructured_horizontal_has_unit_stride,
        )

        if self.auto_optimize:
            auto_optimize_args = {} if self.auto_optimize_args is None else self.auto_optimize_args

            gtx_transformations.gt_auto_optimize(
                sdfg,
                gpu=on_gpu,
                constant_symbols=constant_symbols,
                **auto_optimize_args,
            )
        elif on_gpu:
            # Note that `gt_substitute_compiletime_symbols()` will run `gt_simplify()`
            # at entry, in order to avoid some issue in constant propagatation.
            # Besides, `gt_simplify()` will bring the SDFG into a canonical form
            # that the GPU transformations can handle. This is a workaround for
            # an issue with scalar expressions that are promoted to symbolic expressions
            # and computed on the host (CPU), but the intermediate result is written
            # to a GPU global variable (https://github.com/spcl/dace/issues/1773).
            gtx_transformations.gt_substitute_compiletime_symbols(
                sdfg, constant_symbols, validate=True
            )
            gtx_transformations.gt_gpu_transformation(sdfg, try_removing_trivial_maps=True)

        elif len(constant_symbols) != 0:
            # Target CPU without SDFG transformations, but still replace constant symbols.
            # Replacing the SDFG symbols for field origin in global arrays is strictly
            # required by dace orchestration, which runs the translation stage
            # with `disable_field_origin_on_program_arguments=True`. The program
            # decorator used by dace orchestartion cannot handle field origin.
            # It also requires skipping auto-optimize on the `SDFGConvertible` objects,
            # because it targets the full-application SDFG, so we have to explicitly
            # apply `gt_substitute_compiletime_symbols()` here.
            gtx_transformations.gt_substitute_compiletime_symbols(
                sdfg, constant_symbols, validate=True
            )

        if self.async_sdfg_call:
            make_sdfg_call_async(sdfg, on_gpu)
        else:
            make_sdfg_call_sync(sdfg, on_gpu)

        if self.use_metrics:
            add_instrumentation(sdfg, on_gpu)

        return sdfg

    def __call__(
        self, inp: definitions.CompilableProgramDef
    ) -> stages.ProgramSource[code_specs.SDFGCodeSpec]:
        """Generate DaCe SDFG file from the GTIR definition."""
        program: itir.Program = inp.data
        assert isinstance(program, itir.Program)

        sdfg = self.generate_sdfg(
            program,
            inp.args.offset_provider,  # TODO(havogt): should be offset_provider_type once the transformation don't require run-time info
            inp.args.column_axis,
        )

        arg_types = inp.args.args

        program_parameters = tuple(
            interface.Parameter(param.id, arg_type)
            for param, arg_type in zip(program.params, arg_types)
        )

        module: stages.ProgramSource[code_specs.SDFGCodeSpec] = stages.ProgramSource(
            entry_point=interface.Function(program.id, program_parameters),
            # Set 'hash=True' to compute the SDFG hash and store it in the JSON.
            #   We compute the hash in order to refresh `cfg_list` on the SDFG,
            #   which makes the JSON serialization stable.
            source_code=sdfg.to_json(hash=True),
            library_deps=tuple(),
            code_spec=code_specs.SDFGCodeSpec(),
        )
        return module


class DaCeTranslationStepFactory(factory.Factory):
    class Meta:
        model = DaCeTranslator



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


from gt4py.next.program_processors.runners.dace.lowering.gtir_to_sdfg import build_sdfg_from_gtir
from gt4py.next.program_processors.runners.dace.lowering.gtir_to_sdfg_utils import (
    flatten_tuple_fields,
    get_map_variable,
)


__all__ = [
    "build_sdfg_from_gtir",
    "flatten_tuple_fields",
    "get_map_variable",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from gt4py import storage as gt_storage
from gt4py.cartesian.backend.base import register
from gt4py.cartesian.backend.gtc_common import (
    BackendCodegen,
    BaseGTBackend,
    CUDAPyExtModuleGenerator,
    GTBackendOptions,
    bindings_main_template,
    pybuffer_to_sid,
)
from gt4py.cartesian.gtc.common import DataType
from gt4py.cartesian.gtc.gtcpp import gtcpp, gtcpp_codegen
from gt4py.cartesian.gtc.gtcpp.oir_to_gtcpp import OIRToGTCpp
from gt4py.cartesian.gtc.gtir_to_oir import GTIRToOIR
from gt4py.cartesian.gtc.passes.oir_pipeline import DefaultPipeline
from gt4py.eve import codegen


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_object import StencilObject


class GTExtGenerator(BackendCodegen):
    def __init__(self, class_name: str, module_name: str, backend: BaseGTBackend) -> None:
        self.class_name = class_name
        self.module_name = module_name
        self.backend = backend

    def __call__(self) -> dict[str, dict[str, str]]:
        base_oir = GTIRToOIR().visit(self.backend.builder.gtir)
        oir_pipeline = self.backend.builder.options.backend_opts.get(
            "oir_pipeline", DefaultPipeline()
        )
        oir_node = oir_pipeline.run(base_oir)
        gtcpp_ir = OIRToGTCpp().visit(oir_node)
        format_source = self.backend.builder.options.format_source
        implementation = gtcpp_codegen.GTCppCodegen.apply(
            gtcpp_ir, gt_backend_t=self.backend.GT_BACKEND_T, format_source=format_source
        )
        bindings = GTCppBindingsCodegen.apply(
            gtcpp_ir,
            module_name=self.module_name,
            backend=self.backend,
            format_source=format_source,
        )
        bindings_ext = ".cu" if self.backend.GT_BACKEND_T == "gpu" else ".cpp"
        return {
            "computation": {"computation.hpp": implementation},
            "bindings": {"bindings" + bindings_ext: bindings},
        }


class GTCppBindingsCodegen(codegen.TemplatedGenerator):
    def __init__(self, backend):
        self.backend = backend
        self._unique_index: int = 0

    def unique_index(self) -> int:
        self._unique_index += 1
        return self._unique_index

    def visit_DataType(self, dtype: DataType, **kwargs):
        return gtcpp_codegen.GTCppCodegen().visit_DataType(dtype)

    def visit_FieldDecl(self, node: gtcpp.FieldDecl, **kwargs):
        backend = kwargs["backend"]
        if "external_arg" in kwargs:
            domain_ndim = node.dimensions.count(True)
            data_ndim = len(node.data_dims)
            sid_ndim = domain_ndim + data_ndim
            if kwargs["external_arg"]:
                return "py::{pybind_type} {name}, std::array<gt::int_t,{sid_ndim}> {name}_origin".format(
                    pybind_type=(
                        "object" if self.backend.storage_info["device"] == "gpu" else "buffer"
                    ),
                    name=node.name,
                    sid_ndim=sid_ndim,
                )
            else:
                return pybuffer_to_sid(
                    name=node.name,
                    ctype=self.visit(node.dtype),
                    domain_dim_flags=node.dimensions,
                    data_ndim=len(node.data_dims),
                    stride_kind_index=self.unique_index(),
                    backend=backend,
                )

    def visit_GlobalParamDecl(self, node: gtcpp.GlobalParamDecl, **kwargs):
        if "external_arg" in kwargs:
            if kwargs["external_arg"]:
                return "{dtype} {name}".format(name=node.name, dtype=self.visit(node.dtype))
            else:
                return "gridtools::stencil::global_parameter({name})".format(name=node.name)

    def visit_Program(self, node: gtcpp.Program, **kwargs):
        assert "module_name" in kwargs
        entry_params = self.visit(node.parameters, external_arg=True, **kwargs)
        sid_params = self.visit(node.parameters, external_arg=False, **kwargs)
        return self.generic_visit(node, entry_params=entry_params, sid_params=sid_params, **kwargs)

    Program = bindings_main_template()

    @classmethod
    def apply(cls, root, *, module_name="stencil", **kwargs) -> str:
        generated_code = cls(kwargs.get("backend")).visit(root, module_name=module_name, **kwargs)
        if kwargs.get("format_source", True):
            generated_code = codegen.format_source("cpp", generated_code, style="LLVM")

        return generated_code


class GTBaseBackend(BaseGTBackend):
    options = BaseGTBackend.GT_BACKEND_OPTS
    PYEXT_GENERATOR_CLASS = GTExtGenerator

    def _generate_extension(self, uses_cuda: bool) -> None:
        return self.make_extension(uses_cuda=uses_cuda)

    def generate(self) -> type[StencilObject]:
        self.check_options(self.builder.options)

        # TODO(havogt) add bypass if computation has no effect
        self.generate_extension()

        # Generate and return the Python wrapper class
        return self.make_module()


@register
class GTCpuIfirstBackend(GTBaseBackend):
    """GridTools python backend using gtc."""

    name = "gt:cpu_ifirst"
    GT_BACKEND_T = "cpu_ifirst"
    languages: ClassVar[dict] = {"computation": "c++", "bindings": ["python"]}
    storage_info = gt_storage.layout.CPUIFirstLayout

    def generate_extension(self) -> None:
        return super()._generate_extension(uses_cuda=False)


@register
class GTCpuKfirstBackend(GTBaseBackend):
    """GridTools python backend using gtc."""

    name = "gt:cpu_kfirst"
    GT_BACKEND_T = "cpu_kfirst"
    languages: ClassVar[dict] = {"computation": "c++", "bindings": ["python"]}
    storage_info = gt_storage.layout.CPUKFirstLayout

    def generate_extension(self) -> None:
        return super()._generate_extension(uses_cuda=False)


@register
class GTGpuBackend(GTBaseBackend):
    """GridTools python backend using gtc."""

    MODULE_GENERATOR_CLASS = CUDAPyExtModuleGenerator
    name = "gt:gpu"
    GT_BACKEND_T = "gpu"
    languages: ClassVar[dict] = {"computation": "cuda", "bindings": ["python"]}
    options: ClassVar[GTBackendOptions] = {
        **BaseGTBackend.GT_BACKEND_OPTS,
        "device_sync": {"versioning": True, "type": bool},
    }
    storage_info = gt_storage.layout.GPULayout

    def generate_extension(self) -> None:
        return super()._generate_extension(uses_cuda=True)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import os
import pathlib
import subprocess
import warnings
from typing import TypeVar

from gt4py._core import definitions as core_defs
from gt4py.next import config, errors
from gt4py.next.otf import code_specs, stages
from gt4py.next.otf.compilation import build_data, cache, common, compiler
from gt4py.next.otf.compilation.build_systems import cmake_lists


def get_device_arch() -> str | None:
    if core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.CUDA:
        # use `cp` from core_defs to avoid trying to re-import cupy
        try:
            return core_defs.cp.cuda.Device(0).compute_capability  # type: ignore[attr-defined]
        except core_defs.cp.cuda.runtime.CUDARuntimeError as e:  # type: ignore[attr-defined]
            warnings.warn(
                UserWarning(f"Could not determine the CUDA compute capability: {e}"), stacklevel=2
            )
            return None
    elif core_defs.CUPY_DEVICE_TYPE == core_defs.DeviceType.ROCM:
        # TODO(egparedes): Implement this properly, either parsing the output of `$ rocminfo`
        # or using the HIP low level bindings.
        # Check: https://rocm.docs.amd.com/projects/hip-python/en/latest/user_guide/1_usage.html
        return "gfx942"  # MI300A

    return None


def get_cmake_device_arch_option() -> str:
    cmake_flag = ""
    device_archs: str | None = ""

    match core_defs.CUPY_DEVICE_TYPE:
        case core_defs.DeviceType.CUDA:
            cmake_flag_template = "-DCMAKE_CUDA_ARCHITECTURES={device_archs}"
            device_archs = os.environ.get("CUDAARCHS", "").strip() or get_device_arch()
        case core_defs.DeviceType.ROCM:
            cmake_flag_template = "-DCMAKE_HIP_ARCHITECTURES={device_archs}"
            # `HIPARCHS` is not officially supported by CMake yet, but it might be in the future
            device_archs = os.environ.get("HIPARCHS", "").strip() or get_device_arch()

    cmake_flag = cmake_flag_template.format(device_archs=device_archs) if device_archs else ""

    return cmake_flag


CPPLikeCodeSpecT = TypeVar("CPPLikeCodeSpecT", bound=code_specs.CPPLikeCodeSpec)


@dataclasses.dataclass
class CMakeFactory(
    compiler.BuildSystemProjectGenerator[CPPLikeCodeSpecT, code_specs.PythonCodeSpec]
):
    """Create a CMakeProject from a ``CompilableSource`` stage object with given CMake settings."""

    cmake_generator_name: str = "Ninja"
    cmake_build_type: config.CMakeBuildType = config.CMakeBuildType.DEBUG
    cmake_extra_flags: list[str] = dataclasses.field(default_factory=list)

    def __call__(
        self,
        source: stages.CompilableProject[CPPLikeCodeSpecT, code_specs.PythonCodeSpec],
        cache_lifetime: config.BuildCacheLifetime,
    ) -> CMakeProject:
        if not source.binding_source:
            raise NotImplementedError(
                "CMake build system project requires separate bindings code file."
            )
        name = source.program_source.entry_point.name
        header_name = f"{name}.{source.program_source.code_spec.header_extension}"
        bindings_name = f"{name}_bindings.{source.program_source.code_spec.file_extension}"
        cmake_languages = [cmake_lists.Language(name="CXX")]
        if (src_lang_name := source.program_source.code_spec.source_language) in {
            code_specs.CUDACodeSpec.source_language,
            code_specs.HIPCodeSpec.source_language,
        }:
            cmake_languages = [*cmake_languages, cmake_lists.Language(name=src_lang_name)]
            if device_arch_flag := get_cmake_device_arch_option():
                self.cmake_extra_flags.append(device_arch_flag)
        cmake_lists_src = cmake_lists.generate_cmakelists_source(
            name,
            source.library_deps,
            [header_name, bindings_name],
            languages=cmake_languages,
        )

        return CMakeProject(
            root_path=cache.get_cache_folder(source, cache_lifetime),
            source_files={
                header_name: source.program_source.source_code,
                bindings_name: source.binding_source.source_code,
                "CMakeLists.txt": cmake_lists_src,
            },
            program_name=name,
            generator_name=self.cmake_generator_name,
            build_type=self.cmake_build_type,
            extra_cmake_flags=self.cmake_extra_flags,
        )


@dataclasses.dataclass
class CMakeProject(stages.BuildSystemProject[CPPLikeCodeSpecT, code_specs.PythonCodeSpec]):
    """
    CMake build system for gt4py programs.

    Write source files to disk and run cmake on them, keeping the build data updated.
    The ``.build()`` method runs all the necessary steps to compile the code to a python extension module.
    The location of the module can be found by accessing the ``gt4py.next.otf.compilation.build_data.BuildData``
    on the ``.root_path`` after building.
    """

    root_path: pathlib.Path
    source_files: dict[str, str]
    program_name: str
    generator_name: str = "Ninja"
    build_type: config.CMakeBuildType = config.CMakeBuildType.DEBUG
    extra_cmake_flags: list[str] = dataclasses.field(default_factory=list)

    def build(self) -> None:
        self._write_files()
        self._run_config()
        self._run_build()

    def _write_files(self) -> None:
        for name, content in self.source_files.items():
            (self.root_path / name).write_text(content, encoding="utf-8")

        build_data.write_data(
            build_data.BuildData(
                status=build_data.BuildStatus.INITIALIZED,
                module=pathlib.Path(
                    f"build/bin/{self.program_name}.{common.python_module_suffix()}"
                ),
                entry_point_name=self.program_name,
            ),
            self.root_path,
        )

        self._write_configure_script()

    @property
    def _config_command(self) -> list[str]:
        return [
            "cmake",
            "-G",
            self.generator_name,
            "-S",
            str(self.root_path),
            "-B",
            str(self.root_path / "build"),
            f"-DCMAKE_BUILD_TYPE={self.build_type.value}",
            *self.extra_cmake_flags,
        ]

    def _write_configure_script(self) -> None:
        # TODO(havogt): additionally we could store all `env` vars
        configure_script_path = self.root_path / "configure.sh"
        with configure_script_path.open("w") as build_script_pointer:
            build_script_pointer.write("#!/bin/sh\n")
            build_script_pointer.write(" ".join(self._config_command))
        try:
            configure_script_path.chmod(0o755)
        except OSError:
            # if setting permissions fails, it's not a problem
            pass

    def _run_config(self) -> None:
        logfile = self.root_path / "log_config.txt"
        try:
            with logfile.open(mode="w") as log_file_pointer:
                subprocess.check_call(
                    self._config_command,
                    stdout=log_file_pointer,
                    stderr=log_file_pointer,
                )
        except subprocess.CalledProcessError as e:
            with logfile.open(mode="r") as log_file_pointer:
                log = log_file_pointer.read()
            raise errors.CompilationError(log) from e

        build_data.update_status(new_status=build_data.BuildStatus.CONFIGURED, path=self.root_path)

    def _run_build(self) -> None:
        logfile = self.root_path / "log_build.txt"
        try:
            with logfile.open(mode="w") as log_file_pointer:
                subprocess.check_call(
                    ["cmake", "--build", self.root_path / "build"],
                    stdout=log_file_pointer,
                    stderr=log_file_pointer,
                )
        except subprocess.CalledProcessError as e:
            with logfile.open(mode="r") as log_file_pointer:
                log = log_file_pointer.read()
            raise errors.CompilationError(log) from e

        build_data.update_status(new_status=build_data.BuildStatus.COMPILED, path=self.root_path)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import datetime
import enum
import os
import pathlib
import warnings
from typing import Final


class BuildCacheLifetime(enum.Enum):
    SESSION = 1
    PERSISTENT = 2


class CMakeBuildType(enum.Enum):
    """
    CMake build types enum.

    Member values have to be valid CMake syntax.
    """

    DEBUG = "Debug"
    RELEASE = "Release"
    REL_WITH_DEB_INFO = "RelWithDebInfo"
    MIN_SIZE_REL = "MinSizeRel"


def env_flag_to_bool(name: str, default: bool) -> bool:
    """Convert environment variable string variable to a bool value."""
    flag_value = os.environ.get(name, None)
    if flag_value is None:
        return default
    match flag_value.lower():
        case "0" | "false" | "off":
            return False
        case "1" | "true" | "on":
            return True
        case _:
            raise ValueError(
                "Invalid GT4Py environment flag value: use '0 | false | off' or '1 | true | on'."
            )


def env_flag_to_int(name: str, default: int) -> int:
    """Convert environment variable string variable to an int value."""
    flag_value = os.environ.get(name, None)
    if flag_value is None:
        return default
    try:
        return int(flag_value)
    except ValueError:
        raise ValueError(
            f"Invalid GT4Py environment flag value: {flag_value} is not an integer."
        ) from None


#: Master debug flag
#: Changes defaults for all the other options to be as helpful for debugging as possible.
#: Does not override values set in environment variables.
DEBUG: Final[bool] = env_flag_to_bool("GT4PY_DEBUG", default=False)


#: Verbose flag for DSL compilation errors
VERBOSE_EXCEPTIONS: bool = env_flag_to_bool(
    "GT4PY_VERBOSE_EXCEPTIONS", default=True if DEBUG else False
)


#: Where generated code projects should be persisted.
#: Only active if BUILD_CACHE_LIFETIME is set to PERSISTENT
BUILD_CACHE_DIR: pathlib.Path = (
    pathlib.Path(os.environ.get("GT4PY_BUILD_CACHE_DIR", pathlib.Path.cwd())) / ".gt4py_cache"
)


#: Whether generated code projects should be kept around between runs.
#: - SESSION: generated code projects get destroyed when the interpreter shuts down
#: - PERSISTENT: generated code projects are written to BUILD_CACHE_DIR and persist between runs
BUILD_CACHE_LIFETIME: BuildCacheLifetime = BuildCacheLifetime[
    os.environ.get("GT4PY_BUILD_CACHE_LIFETIME", "persistent" if DEBUG else "session").upper()
]

#: Build type to be used when CMake is used to compile generated code.
#: Might have no effect when CMake is not used as part of the toolchain.
# FIXME[#2447](egparedes): compile-time setting, should be included in the build cache key.
CMAKE_BUILD_TYPE: CMakeBuildType = CMakeBuildType[
    os.environ.get("GT4PY_CMAKE_BUILD_TYPE", "debug" if DEBUG else "release").upper()
]

#: Experimental, use at your own risk: assume horizontal dimension has stride 1
# FIXME[#2447](egparedes): compile-time setting, should be included in the build cache key.
UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE: bool = env_flag_to_bool(
    "GT4PY_UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE", default=False
)

#: Add GPU trace markers (NVTX, ROC-TX) to the generated code, at compile time.
# FIXME[#2447](egparedes): compile-time setting, should be included in the build cache key.
ADD_GPU_TRACE_MARKERS: bool = env_flag_to_bool("GT4PY_ADD_GPU_TRACE_MARKERS", default=False)

#: Number of threads to use to use for compilation (0 = synchronous compilation).
#: Default:
#: - use os.cpu_count(), TODO(havogt): in Python >= 3.13 use `process_cpu_count()`
#: - if os.cpu_count() is None we are conservative and use 1 job,
#: - if the number is huge (e.g. HPC system) we limit to a smaller number
BUILD_JOBS: int = int(os.environ.get("GT4PY_BUILD_JOBS", min(os.cpu_count() or 1, 32)))

#: User-defined level to enable metrics at lower or equal level.
#: Enabling metrics collection will do extra synchronization and will have
#: impact on runtime performance.
COLLECT_METRICS_LEVEL: int = env_flag_to_int("GT4PY_COLLECT_METRICS_LEVEL", default=0)


#: File path to dump collected metrics at exit, if COLLECT_METRICS_LEVEL is enabled.
#: If set to a True value, it defaults to "gt4py_metrics_YYYYMMDD_HHMMSS.json" in
#: the current folder.
DUMP_METRICS_AT_EXIT: str | None = None


#: Filter out DaCe related warnings. If not set warnings will be suppressed if the
#: code runs in no debug mode.
SKIP_DACE_WARNINGS: bool = env_flag_to_bool("GT4PY_SKIP_DACE_WARNINGS", default=not __debug__)


if SKIP_DACE_WARNINGS:
    # NOTE: Ideally we would suppress the warnings using context managers directly in
    #   the backend. However, because this is not thread safe in Python versions before
    #   3.14, we have to do it here.
    warnings.filterwarnings(action="ignore", module="^dace(\..+)?")
    warnings.filterwarnings(
        action="ignore", module="^gt4py.next.program_processors.runners.dace.transformations(\..+)?"
    )


def _init_dump_metrics_filename() -> str:
    return f"gt4py_metrics_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.json"


_dump_metrics_at_exit_env = os.environ.get("GT4PY_DUMP_METRICS_AT_EXIT", None)
if _dump_metrics_at_exit_env is not None:
    try:
        if env_flag_to_bool("GT4PY_DUMP_METRICS_AT_EXIT", default=False):
            DUMP_METRICS_AT_EXIT = _init_dump_metrics_filename()
    except ValueError:
        DUMP_METRICS_AT_EXIT = _dump_metrics_at_exit_env


#: The default for whether to allow jit-compilation for a compiled program.
#: This default can be overriden per program.
ENABLE_JIT_DEFAULT: bool = env_flag_to_bool("GT4PY_ENABLE_JIT_DEFAULT", default=True)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
from typing import Any

from gt4py.cartesian import utils as gt_utils
from gt4py.cartesian.definitions import BuildOptions, StencilID
from gt4py.cartesian.gtc import gtir
from gt4py.cartesian.type_hints import AnnotatedStencilFunc, AnyStencilFunc


class Frontend(abc.ABC):
    name: str
    """Frontend name."""

    @classmethod
    @abc.abstractmethod
    def get_stencil_id(
        cls,
        qualified_name: str,
        definition: AnyStencilFunc,
        externals: dict[str, Any],
        options_id: str,
    ) -> StencilID:
        """
        Create a StencilID object that contains a unique hash for the stencil.

        Notes
        -----
        This method seems to no longer be used through StencilBuilder.

        Returns
        -------
        StencilID:
            An object that contains the qualified name and unique hash.

        Raises
        ------
        GTSyntaxError
            If there is a parsing error.

        TypeError
            If there is a error resolving external types.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def generate(
        cls,
        definition: AnyStencilFunc,
        externals: dict[str, Any],
        dtypes: dict[type, type],
        options: BuildOptions,
        backend_name: str,
    ) -> gtir.Stencil:
        """
        Generate a StencilDefinition from a stencil Python function.

        Raises
        ------
        GTScriptSyntaxError
            If there is a parsing error.

        GTScriptDefinitionError
            If types are misused in the definition.

        TypeError
            If there is an error in resolving external types.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def prepare_stencil_definition(
        cls, definition: AnyStencilFunc, externals: dict[str, Any]
    ) -> AnnotatedStencilFunc:
        """
        Annotate the stencil function if not already done so.

        Raises
        ------
        GTSyntaxError
            If there is a parsing error.

        TypeError
            If there is a error resolving external types.
        """
        pass


REGISTRY = gt_utils.Registry[type[Frontend]]()


def from_name(name: str) -> type[Frontend]:
    """Return frontend by name."""
    frontend_cls = REGISTRY.get(name, None)
    if frontend_cls is None:
        raise ValueError(
            f"Frontend '{name} is not registered. Valid options are: {REGISTRY.names}."
        )
    return frontend_cls


def register(frontend_cls: type[Frontend]) -> type[Frontend]:
    """Register a new frontend."""
    return REGISTRY.register(frontend_cls.name, frontend_cls)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Python bindings generator for C++ functions."""

from __future__ import annotations

from collections.abc import Collection
from typing import Any, Optional, Sequence, TypeVar, Union

import gt4py.eve as eve
from gt4py.eve.codegen import JinjaTemplate as as_jinja, TemplatedGenerator
from gt4py.next import common, config
from gt4py.next.otf import code_specs, cpp_utils, stages, workflow
from gt4py.next.otf.binding import cpp_interface, interface
from gt4py.next.type_system import type_specifications as ts


CodeSpecT = TypeVar("CodeSpecT", bound=code_specs.CPPLikeCodeSpec, covariant=True)


class Expr(eve.Node):
    pass


class DimensionSpec(Expr):
    name: str
    static_stride: Optional[int]


class BufferSID(Expr):
    source_buffer: str
    dimensions: Sequence[DimensionSpec]
    scalar_type: ts.ScalarType


class Tuple(Expr):
    elems: list[Union[Expr, str]]


class FunctionCall(Expr):
    target: interface.Function
    args: Sequence[Any]


class ReturnStmt(eve.Node):
    expr: Expr


class ExprStmt(eve.Node):
    expr: Expr


class FunctionParameter(eve.Node):
    name: str
    type_: ts.TypeSpec


class WrapperFunction(eve.Node):
    name: str
    parameters: Sequence[FunctionParameter]
    body: Union[ExprStmt, ReturnStmt]
    on_device: bool = False


class BindingFunction(eve.Node):
    exported_name: str
    wrapper_name: str
    doc: str
    n_params: int


class BindingModule(eve.Node):
    name: str
    doc: str
    functions: Sequence[BindingFunction]


class BindingFile(eve.Node):
    callee_header_file: str
    header_files: Sequence[str]
    wrapper: WrapperFunction
    binding_module: BindingModule


def _type_string(type_: ts.TypeSpec) -> str:
    if isinstance(type_, ts.TupleType):
        return f"std::tuple<{','.join(_type_string(t) for t in type_.types)}>"
    elif isinstance(type_, ts.FieldType):
        ndims = len(type_.dims)
        # cannot be ListType: the concept is represented as Field with local Dimension in this interface
        assert isinstance(type_.dtype, ts.ScalarType)
        dtype = cpp_utils.pytype_to_cpptype(type_.dtype)
        shape = f"nanobind::shape<{', '.join(['gridtools::nanobind::dynamic_size'] * ndims)}>"
        buffer_t = f"nanobind::ndarray<{dtype}, {shape}>"
        origin_t = f"std::tuple<{', '.join(['ptrdiff_t'] * ndims)}>"
        return f"std::pair<{buffer_t}, {origin_t}>"
    elif isinstance(type_, ts.ScalarType):
        return cpp_utils.pytype_to_cpptype(type_)
    else:
        raise ValueError(f"Type '{type_}' is not supported in nanobind interfaces.")


class BindingCodeGenerator(TemplatedGenerator):
    BindingFile = as_jinja(
        """\
        #include "{{callee_header_file}}"

        {% for header_file in header_files: %}\
        #include <{{header_file}}>
        {% endfor %}

        {{wrapper}}

        {{binding_module}}\
        """
    )

    def visit_WrapperFunction(
        self, node: WrapperFunction, **kwargs: Any
    ) -> Union[str, Collection[str]]:
        return_stmt = "return _gt4py_return;" if isinstance(node.body, ReturnStmt) else ""
        return self.generic_visit(node, return_stmt=return_stmt)

    WrapperFunction = as_jinja(
        """\
        decltype(auto) {{name}}(
            {{"\n,".join(parameters + ["std::optional<nanobind::dict> exec_info"])}}
        )
        {
            if (!exec_info.has_value()) {
                {{body}}            
                {{return_stmt}}
            }
            else {
                auto start = std::chrono::high_resolution_clock::now();
                {{body}}
                {% if _this_node.on_device %}cudaDeviceSynchronize();{% endif %}
                auto stop = std::chrono::high_resolution_clock::now();
                exec_info->operator[]("run_cpp_duration") = std::chrono::duration<double>(stop - start).count();
                {{return_stmt}}
            }
        }\
        """
    )

    FunctionParameter = as_jinja("{{_this_module._type_string(_this_node.type_)}} {{name}}")

    ExprStmt = as_jinja("""{{expr}};""")

    ReturnStmt = as_jinja("""auto _gt4py_return = {{expr}};""")

    BindingModule = as_jinja(
        """\
        NB_MODULE({{name}}, module) {
            module.doc() = "{{doc}}";
            {{"\n".join(functions)}}
        }\
        """
    )

    BindingFunction = as_jinja(
        """module.def("{{exported_name}}", &{{wrapper_name}}, "{{doc}}", {{", ".join(["nanobind::arg()"] * _this_node.n_params + ['nanobind::arg("exec_info") = nanobind::none()'])}});"""
    )

    def visit_FunctionCall(self, call: FunctionCall, **kwargs: Any) -> str:
        args = [self.visit(arg) for arg in call.args]
        return cpp_interface.render_function_call(call.target, args)

    def visit_BufferSID(self, sid: BufferSID, **kwargs: Any) -> str:
        pybuffer = f"{sid.source_buffer}.first"
        dims = [self.visit(dim) for dim in sid.dimensions]
        origin = f"{sid.source_buffer}.second"
        stride_spec = [
            "gridtools::nanobind::dynamic_size"
            if dim.static_stride is None
            else str(dim.static_stride)
            for dim in sid.dimensions
        ]
        stride_spec_string = (
            f"gridtools::nanobind::stride_spec<{', '.join(str(s) for s in stride_spec)}>{{}}"
        )

        as_sid = f"gridtools::nanobind::as_sid({pybuffer}, {stride_spec_string})"
        shifted = f"gridtools::sid::shift_sid_origin({as_sid}, {origin})"
        renamed = f"gridtools::sid::rename_numbered_dimensions<{', '.join(dims)}>({shifted})"
        return renamed

    Tuple = as_jinja("""gridtools::tuple({{','.join(elems)}})""")

    DimensionSpec = as_jinja("""generated::{{name}}_t""")


def _tuple_get(index: int, var: str) -> str:
    return f"gridtools::tuple_util::get<{index}>({var})"


def make_argument(name: str, type_: ts.TypeSpec) -> str | BufferSID | Tuple:
    if isinstance(type_, ts.FieldType):
        return BufferSID(
            source_buffer=name,
            dimensions=[
                DimensionSpec(
                    name=dim.value,
                    static_stride=1
                    if (
                        config.UNSTRUCTURED_HORIZONTAL_HAS_UNIT_STRIDE
                        and dim.kind == common.DimensionKind.HORIZONTAL
                    )
                    else None,
                )
                for dim in type_.dims
            ],
            scalar_type=type_.dtype,
        )
    elif isinstance(type_, ts.TupleType):
        elements = [make_argument(_tuple_get(i, name), t) for i, t in enumerate(type_.types)]
        return Tuple(elems=elements)
    elif isinstance(type_, ts.ScalarType):
        return name
    else:
        raise ValueError(f"Type '{type_}' is not supported in nanobind interfaces.")


def create_bindings(
    program_source: stages.ProgramSource[CodeSpecT],
) -> stages.BindingSource[CodeSpecT, code_specs.PythonCodeSpec]:
    """
    Generate Python bindings through which a C++ function can be called.

    Parameters
    ----------
    program_source
        The program source for which the bindings are created
    """
    if not isinstance(program_source.code_spec, code_specs.CPPLikeCodeSpec):
        raise ValueError(
            f"Can only create bindings for C++ program sources, received '{program_source.code_spec.source_language}'."
        )
    wrapper_name = program_source.entry_point.name + "_wrapper"

    Stmt = ReturnStmt if program_source.entry_point.returns else ExprStmt
    file_binding = BindingFile(
        callee_header_file=f"{program_source.entry_point.name}.{program_source.code_spec.header_extension}",
        header_files=[
            "chrono",
            "optional",
            "nanobind/nanobind.h",
            "nanobind/stl/tuple.h",
            "nanobind/stl/pair.h",
            "nanobind/stl/optional.h",
            "nanobind/ndarray.h",
            "gridtools/sid/composite.hpp",
            "gridtools/sid/unknown_kind.hpp",
            "gridtools/sid/rename_dimensions.hpp",
            "gridtools/common/defs.hpp",
            "gridtools/common/tuple_util.hpp",
            "gridtools/fn/unstructured.hpp",
            "gridtools/fn/cartesian.hpp",
            "gridtools/storage/adapter/nanobind_adapter.hpp",
        ],
        wrapper=WrapperFunction(
            name=wrapper_name,
            parameters=[
                FunctionParameter(name=param.name, type_=param.type_)
                for param in program_source.entry_point.parameters
            ],
            body=Stmt(
                expr=FunctionCall(
                    target=program_source.entry_point,
                    args=[
                        make_argument(param.name, param.type_)
                        for param in program_source.entry_point.parameters
                    ],
                )
            ),
            on_device=isinstance(
                program_source.code_spec,
                (code_specs.CUDACodeSpec, code_specs.HIPCodeSpec),
            ),
        ),
        binding_module=BindingModule(
            name=program_source.entry_point.name,
            doc="",
            functions=[
                BindingFunction(
                    exported_name=program_source.entry_point.name,
                    wrapper_name=wrapper_name,
                    doc="",
                    n_params=len(program_source.entry_point.parameters),
                )
            ],
        ),
    )

    src = interface.format_source(
        program_source.code_spec, BindingCodeGenerator.apply(file_binding)
    )

    return stages.BindingSource(src, (interface.LibraryDependency("nanobind", "2.0.0"),))


@workflow.make_step
def bind_source(
    inp: stages.ProgramSource[CodeSpecT],
) -> stages.CompilableProject[CodeSpecT, code_specs.PythonCodeSpec]:
    return stages.CompilableProject(program_source=inp, binding_source=create_bindings(inp))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Annotated, Final, TypeAlias

from gt4py._core.definitions import Scalar
from gt4py.next import backend, common, constructors
from gt4py.next.ffront import decorator, stages as ffront_stages
from gt4py.next.otf import compiled_program


_ONLY_FOR_TYPING: Final[str] = "only for typing"

# TODO(havogt): alternatively we could introduce Protocols
DSLDefinition: TypeAlias = Annotated[ffront_stages.DSLDefinition, _ONLY_FOR_TYPING]

Program: TypeAlias = Annotated[decorator.Program, _ONLY_FOR_TYPING]
FieldOperator: TypeAlias = Annotated[decorator.FieldOperator, _ONLY_FOR_TYPING]
GTEntryPoint: TypeAlias = Annotated[decorator.GTEntryPoint, _ONLY_FOR_TYPING]

CompiledProgramsKey: TypeAlias = Annotated[compiled_program.CompiledProgramsKey, _ONLY_FOR_TYPING]

Backend: TypeAlias = Annotated[backend.Backend, _ONLY_FOR_TYPING]

Allocator: TypeAlias = Annotated[constructors.Allocator, _ONLY_FOR_TYPING]

OffsetProvider: TypeAlias = Annotated[common.OffsetProvider, _ONLY_FOR_TYPING]


__all__ = [
    "Allocator",
    "Backend",
    "FieldOperator",
    "OffsetProvider",
    "Program",
    # from _core.definitions for convenience
    "Scalar",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.cartesian.gtc import gtir


def prune_unused_parameters(node: gtir.Stencil) -> gtir.Stencil:
    """
    Remove unused parameters from the gtir signature.

    (Maybe this pass should go into a later stage. If you need to touch this pass,
    e.g. when the definition_ir gets removed, consider moving it to a more appropriate
    level. Maybe to the backend IR?)
    """
    assert isinstance(node, gtir.Stencil)
    used_variables = (
        node.walk_values()
        .if_isinstance(gtir.FieldAccess, gtir.ScalarAccess)
        .getattr("name")
        .to_list()
    )
    used_params = list(filter(lambda param: param.name in used_variables, node.params))
    return node.copy(update={"params": used_params})



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Implementation of GTScript: an embedded DSL in Python for stencil computations.

Interface functions to define and compile GTScript definitions and empty symbol
definitions for the keywords of the DSL.
"""

import sys

from gt4py.cartesian.gtscript import *  # noqa: F403 [undefined-local-with-import-star]


sys.modules["__gtscript__"] = sys.modules["gt4py.cartesian.__gtscript__"]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py import eve
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm
from gt4py.next.iterator.transforms import inline_lambdas
from gt4py.next.iterator.type_system import inference as itir_inference
from gt4py.next.type_system import type_specifications as ts


class InlineScalar(eve.NodeTranslator):
    PRESERVED_ANNEX_ATTRS = ("domain",)

    @classmethod
    def apply(cls, program: itir.Program, offset_provider_type: common.OffsetProviderType):
        program = itir_inference.infer(program, offset_provider_type=offset_provider_type)
        return cls().visit(program)

    def generic_visit(self, node, **kwargs):
        if cpm.is_call_to(node, "as_fieldop"):
            return node

        return super().generic_visit(node, **kwargs)

    def visit_Expr(self, node: itir.Expr):
        node = self.generic_visit(node)

        if cpm.is_let(node):
            eligible_params = [isinstance(arg.type, ts.ScalarType) for arg in node.args]
            node = inline_lambdas.inline_lambda(node, eligible_params=eligible_params)
            return node
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from typing import Optional, Protocol

from gt4py.next import common, utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, ir_makers as im
from gt4py.next.iterator.transforms import (
    concat_where,
    dead_code_elimination,
    fuse_as_fieldop,
    global_tmps,
    infer_domain,
    infer_domain_ops,
    inline_dynamic_shifts,
    inline_fundefs,
    inline_lifts,
    prune_empty_concat_where,
    remove_broadcast,
    symbol_ref_utils,
)
from gt4py.next.iterator.transforms.collapse_list_get import CollapseListGet
from gt4py.next.iterator.transforms.collapse_tuple import CollapseTuple
from gt4py.next.iterator.transforms.constant_folding import ConstantFolding
from gt4py.next.iterator.transforms.cse import CommonSubexpressionElimination
from gt4py.next.iterator.transforms.fuse_maps import FuseMaps
from gt4py.next.iterator.transforms.inline_lambdas import InlineLambdas
from gt4py.next.iterator.transforms.inline_scalar import InlineScalar
from gt4py.next.iterator.transforms.merge_let import MergeLet
from gt4py.next.iterator.transforms.normalize_shifts import NormalizeShifts
from gt4py.next.iterator.transforms.unroll_reduce import UnrollReduce
from gt4py.next.iterator.type_system.inference import infer


class GTIRTransform(Protocol):
    def __call__(
        self, _: itir.Program, *, offset_provider: common.OffsetProvider
    ) -> itir.Program: ...


def _max_domain_range_sizes(offset_provider: common.OffsetProvider) -> dict[str, itir.Literal]:
    """
    Extract horizontal domain sizes from an `offset_provider`.

    Considers the shape of the neighbor table to get the size of each `source_dim` and the maximum
    value inside the neighbor table to get the size of each `codomain`.
    """
    sizes: dict[str, int] = {}
    for provider in offset_provider.values():
        if common.is_neighbor_connectivity(provider):
            src_dim = provider.__gt_type__().source_dim.value
            codomain_dim = provider.__gt_type__().codomain.value
            sizes[src_dim] = max(sizes.get(src_dim, 0), provider.ndarray.shape[0])
            sizes[codomain_dim] = max(
                sizes.get(codomain_dim, 0),
                int(provider.ndarray.max()) + 1,  # type: ignore[attr-defined] # TODO(havogt): improve typing for NDArrayObject
            )

    sizes_exprs = {k: im.literal_from_value(v) for k, v in sizes.items()}
    return sizes_exprs


def _has_dynamic_domains(ir: itir.Program) -> bool:
    # note: this function does not respect symbol collisions with builtins. As it is a temporary
    # workaround we don't care about this corner case.
    domains = set()
    domains |= ir.walk_values().if_isinstance(itir.SetAt).getattr("domain").to_set()
    for as_fop in (
        ir.walk_values()
        .if_isinstance(itir.FunCall)
        .filter(lambda node: cpm.is_call_to(node, "as_fieldop") and len(node.args) == 2)
    ):
        domains.add(as_fop.args[1])
    return len(symbol_ref_utils.collect_symbol_refs(domains)) > 0


def _process_symbolic_domains_option(
    ir: itir.Program,
    offset_provider: common.OffsetProvider,
    symbolic_domain_sizes: Optional[dict[str, itir.Expr]],
    use_max_domain_range_on_unstructured_shift: Optional[bool],
) -> Optional[dict[str, itir.Expr]]:
    """
    Given a program, offset_provider and some configuration options determine how domains are
    inferred.

    The output of this function is used as `symbolic_domain_sizes` argument of domain inference, i.e.
    :func:`infer_domain.infer_program`.

    Right now domains of `as_fieldop` expressions can be inferred either a) using static information
    from the offset provider, or b) they are set to an expression controlled by
    the user and configured in the backend, or c) they are set to the maximum possible domain /
    everywhere (see :func:`_max_domain_range_sizes`)

    Option a) applies when the program is decorated with `static_domains = True` (unless option c)
    is explicitly requested). Then all dynamic domains were replaced with static ones
    which we recognize here. The domain inference then uses this static information which we
    communicate by returning `None`, i.e. no symbolic domain sizes.
    Option b) applies when the user explicitly configured `symbolic_domain_sizes` in the backend.
    In that case we just forward the value.
    Option c) applies when `static_domains = False` or when explicitly configured in the backend
    with `use_max_domain_range_on_unstructured_shift = True`. In that case we determine the
    maximum sizes using :func:`_max_domain_range_sizes` and return them.
    """
    if symbolic_domain_sizes:
        assert not use_max_domain_range_on_unstructured_shift, "Options are mutually exclusive."
        return symbolic_domain_sizes

    if use_max_domain_range_on_unstructured_shift is None:
        use_max_domain_range_on_unstructured_shift = _has_dynamic_domains(ir)
    if use_max_domain_range_on_unstructured_shift:
        # TODO(havogt): ICON4Py uses this codepath as default for now. Once we use the minimal domain range, we should re-enable this warning.
        # if not _has_dynamic_domains(ir):
        #     warnings.warn(
        #         "You are using static domains together with "
        #         "'use_max_domain_range_on_unstructured_shift'. This is "
        #         "likely not what you wanted.",
        #         stacklevel=2,  # noqa: ERA001
        #     )  # noqa: ERA001, RUF100
        assert not symbolic_domain_sizes, "Options are mutually exclusive."
        symbolic_domain_sizes = _max_domain_range_sizes(offset_provider)  # type: ignore[assignment]
    return symbolic_domain_sizes


# TODO(tehrengruber): Revisit interface to configure temporary extraction. We currently forward
#  `extract_temporaries` and `temporary_extraction_heuristics` which is inconvenient.
def apply_common_transforms(
    ir: itir.Program,
    *,
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    extract_temporaries=False,
    unroll_reduce=False,
    common_subexpression_elimination=True,
    force_inline_lambda_args=False,
    #: A dictionary mapping axes names to their length. See :func:`infer_domain.infer_expr` for
    #: more details.
    symbolic_domain_sizes: Optional[dict[str, itir.Expr]] = None,
    # TODO(tehrengruber): Remove this option again as soon as we have the necessary builtins
    #  to work with / translate domains.
    use_max_domain_range_on_unstructured_shift: Optional[bool] = None,
) -> itir.Program:
    assert isinstance(ir, itir.Program)
    # TODO(tehrengruber): Allow `common.OffsetProviderType`, but domain inference currently
    #  relies on static information or `symbolic_domain_sizes`.
    assert common.is_offset_provider(offset_provider)

    offset_provider_type = common.offset_provider_to_type(offset_provider)

    symbolic_domain_sizes = _process_symbolic_domains_option(
        ir, offset_provider, symbolic_domain_sizes, use_max_domain_range_on_unstructured_shift
    )

    uids = utils.IDGeneratorPool()

    ir = MergeLet().visit(ir)
    ir = inline_fundefs.InlineFundefs().visit(ir)

    ir = inline_fundefs.prune_unreferenced_fundefs(ir)
    ir = NormalizeShifts().visit(ir)

    # TODO(tehrengruber): Many iterator test contain lifts that need to be inlined, e.g.
    #  test_can_deref. We didn't notice previously as FieldOpFusion did this implicitly everywhere.
    ir = inline_lifts.InlineLifts().visit(ir)

    ir = concat_where.expand_tuple_args(ir, offset_provider_type=offset_provider_type)  # type: ignore[assignment]  # always an itir.Program
    ir = dead_code_elimination.dead_code_elimination(
        ir, uids=uids, offset_provider_type=offset_provider_type
    )  # domain inference does not support dead-code
    ir = inline_dynamic_shifts.InlineDynamicShifts.apply(
        ir, offset_provider_type=offset_provider_type, uids=uids
    )  # domain inference does not support dynamic offsets yet
    ir = infer_domain_ops.InferDomainOps.apply(ir)
    ir = concat_where.canonicalize_domain_argument(ir)

    ir = infer_domain.infer_program(
        ir,
        offset_provider=offset_provider,
        symbolic_domain_sizes=symbolic_domain_sizes,
    )
    ir = prune_empty_concat_where.prune_empty_concat_where(ir)
    ir = remove_broadcast.RemoveBroadcast.apply(ir)

    ir = concat_where.transform_to_as_fieldop(ir)

    for _ in range(10):
        inlined = ir

        inlined = InlineLambdas.apply(inlined, opcount_preserving=True)
        inlined = ConstantFolding.apply(inlined)  # type: ignore[assignment]  # always an itir.Program
        # This pass is required to be in the loop such that when an `if_` call with tuple arguments
        # is constant-folded the surrounding tuple_get calls can be removed.
        inlined = CollapseTuple.apply(
            inlined,
            enabled_transformations=~CollapseTuple.Transformation.PROPAGATE_TO_IF_ON_TUPLES,
            uids=uids,
            offset_provider_type=offset_provider_type,
        )  # type: ignore[assignment]  # always an itir.Program
        inlined = InlineScalar.apply(inlined, offset_provider_type=offset_provider_type)

        # This pass is required to run after CollapseTuple as otherwise we can not inline
        # expressions like `tuple_get(make_tuple(as_fieldop(stencil)(...)))` where stencil returns
        # a list. Such expressions must be inlined however because no backend supports such
        # field operators right now.
        inlined = fuse_as_fieldop.FuseAsFieldOp.apply(
            inlined, uids=uids, offset_provider_type=offset_provider_type
        )

        if inlined == ir:
            break
        ir = inlined
    else:
        raise RuntimeError("Inlining 'lift' and 'lambdas' did not converge.")

    # breaks in test_zero_dim_tuple_arg as trivial tuple_get is not inlined
    if common_subexpression_elimination:
        ir = CommonSubexpressionElimination.apply(
            ir, offset_provider_type=offset_provider_type, uids=uids
        )
        ir = MergeLet().visit(ir)
        ir = InlineLambdas.apply(ir, opcount_preserving=True)

    if extract_temporaries:
        ir = infer(ir, inplace=True, offset_provider_type=offset_provider_type)
        ir = global_tmps.create_global_tmps(
            ir,
            offset_provider=offset_provider,
            symbolic_domain_sizes=symbolic_domain_sizes,
            uids=uids,
        )

    ir = NormalizeShifts().visit(ir)

    ir = FuseMaps(uids=uids).visit(ir)
    ir = CollapseListGet().visit(ir)

    if unroll_reduce:
        for _ in range(10):
            unrolled = UnrollReduce.apply(ir, offset_provider_type=offset_provider_type, uids=uids)
            unrolled = CollapseListGet().visit(unrolled)
            unrolled = NormalizeShifts().visit(unrolled)
            # this is required as nested neighbor reductions can contain lifts, e.g.,
            # `neighbors(V2Eₒ, ↑f(...))`
            unrolled = inline_lifts.InlineLifts().visit(unrolled)
            unrolled = NormalizeShifts().visit(unrolled)
            if unrolled == ir:
                break
            ir = unrolled
        else:
            raise RuntimeError("Reduction unrolling failed.")

    ir = InlineLambdas.apply(
        ir, opcount_preserving=True, force_inline_lambda_args=force_inline_lambda_args
    )

    assert isinstance(ir, itir.Program)
    return ir


def apply_fieldview_transforms(
    ir: itir.Program,
    *,
    offset_provider: common.OffsetProvider,
    # TODO(tehrengruber): Remove this option again as soon as we have the necessary builtins
    #  to work with / translate domains.
    use_max_domain_range_on_unstructured_shift: Optional[bool] = None,
) -> itir.Program:
    offset_provider_type = common.offset_provider_to_type(offset_provider)

    uids = utils.IDGeneratorPool()

    symbolic_domain_sizes = _process_symbolic_domains_option(
        ir, offset_provider, None, use_max_domain_range_on_unstructured_shift
    )

    ir = inline_fundefs.InlineFundefs().visit(ir)
    ir = inline_fundefs.prune_unreferenced_fundefs(ir)
    # required for dead-code-elimination and `prune_empty_concat_where` pass
    ir = concat_where.expand_tuple_args(ir, offset_provider_type=offset_provider_type)  # type: ignore[assignment]  # always an itir.Program
    ir = dead_code_elimination.dead_code_elimination(
        ir, offset_provider_type=offset_provider_type, uids=uids
    )
    ir = inline_dynamic_shifts.InlineDynamicShifts.apply(
        ir, offset_provider_type=offset_provider_type, uids=uids
    )  # domain inference does not support dynamic offsets yet

    ir = infer_domain_ops.InferDomainOps.apply(ir)
    ir = concat_where.canonicalize_domain_argument(ir)
    ir = ConstantFolding.apply(ir)  # type: ignore[assignment]  # always an itir.Program

    ir = infer_domain.infer_program(
        ir,
        symbolic_domain_sizes=symbolic_domain_sizes,
        offset_provider=offset_provider,
    )
    ir = prune_empty_concat_where.prune_empty_concat_where(ir)
    ir = remove_broadcast.RemoveBroadcast.apply(ir)
    return ir
